"""
Flow:
1. We need to start with a list of mth5 files.
- Assert that synthetic data exist, (and build if they dont)
- should already know what you expect here ...  test1,2,3.h5 and test12rr.h5
2. Two ways to prepare the FC-schema (processing configs)
- a) use the mt_metadata processing fourier_coefficients structures explictly
- b) use the default processing configs you already use for processing, and
extract type (a) cfgs from these (the machinery to do this should exist already)
3. Loop over files and generate FCs
4. Compare fc values against some archived values

ToDo:
1. Make one test generate the decimation_and_stft_config with default values from
the decimation_and_stft_config_creator method here
2. Make another test take the existing aurora processing config and transform it to
decimation_and_stft_config


Here are the synthetic files for which this is currently passing tests
    [PosixPath('/home/kkappler/software/irismt/aurora/tests/synthetic/data/test1.h5'),
     PosixPath('/home/kkappler/software/irismt/aurora/tests/synthetic/data/test2.h5'),
     PosixPath('/home/kkappler/software/irismt/aurora/tests/synthetic/data/test3.h5'),
     PosixPath('/home/kkappler/software/irismt/aurora/tests/synthetic/data/test12rr.h5')]

"""
import unittest

from aurora.config.config_creator import ConfigCreator
from aurora.pipelines.fourier_coefficients import add_fcs_to_mth5
from aurora.pipelines.fourier_coefficients import decimation_and_stft_config_creator
from aurora.pipelines.fourier_coefficients import read_back_fcs
from aurora.pipelines.process_mth5 import process_mth5
from aurora.pipelines.run_summary import RunSummary
from aurora.test_utils.synthetic.make_mth5_from_asc import create_test1_h5
from aurora.test_utils.synthetic.make_mth5_from_asc import create_test2_h5
from aurora.test_utils.synthetic.make_mth5_from_asc import create_test3_h5
from aurora.test_utils.synthetic.make_mth5_from_asc import create_test12rr_h5
from aurora.test_utils.synthetic.make_processing_configs import create_test_run_config
from aurora.test_utils.synthetic.paths import AURORA_RESULTS_PATH
from aurora.transfer_function.kernel_dataset import KernelDataset
from mth5.helpers import close_open_files


class TestAddFourierCoefficientsToSyntheticData(unittest.TestCase):
    """
    Runs several synthetic processing tests from config creation to tf_cls.

    """

    @classmethod
    def setUpClass(self):
        print("make synthetic data")
        close_open_files()
        self.file_version = "0.1.0"
        mth5_path_1 = create_test1_h5(file_version=self.file_version)
        mth5_path_2 = create_test2_h5(file_version=self.file_version)
        mth5_path_3 = create_test3_h5(file_version=self.file_version)
        mth5_path_12rr = create_test12rr_h5(file_version=self.file_version)
        self.mth5_paths = [mth5_path_1, mth5_path_2, mth5_path_3, mth5_path_12rr]

    def test_123(self):
        """
        This test adds FCs to each of the synthetic files that get built in setUpClass method
        Returns
        -------

        """
        for mth5_path in self.mth5_paths:
            mth5_paths = [
                mth5_path,
            ]
            run_summary = RunSummary()
            run_summary.from_mth5s(mth5_paths)
            tfk_dataset = KernelDataset()

            # Get Processing Config
            if mth5_path.stem in [
                "test1",
                "test2",
            ]:
                station_id = mth5_path.stem
                tfk_dataset.from_run_summary(run_summary, station_id)
                processing_config = create_test_run_config(station_id, tfk_dataset)
            elif mth5_path.stem in [
                "test3",
            ]:
                station_id = "test3"
                tfk_dataset.from_run_summary(run_summary, station_id)
                cc = ConfigCreator()
                processing_config = cc.create_from_kernel_dataset(tfk_dataset)
            elif mth5_path.stem in [
                "test12rr",
            ]:
                tfk_dataset.from_run_summary(run_summary, "test1", "test2")
                cc = ConfigCreator()
                processing_config = cc.create_from_kernel_dataset(tfk_dataset)

            # Extract FC decimations from processing config and build the layer
            fc_decimations = [
                x.to_fc_decimation() for x in processing_config.decimations
            ]
            add_fcs_to_mth5(mth5_path, decimation_and_stft_configs=fc_decimations)
            read_back_fcs(mth5_path)

            # Confirm the file still processes fine with the fcs inside
            tfc = process_mth5(processing_config, tfk_dataset=tfk_dataset)
            return tfc
        return

    def test_decimation_and_stft_config_creator(self):
        cfgs = decimation_and_stft_config_creator(1.0)
        return cfgs

    def test_create_then_use_stored_fcs_for_processing(self):
        """"""
        from test_processing import process_synthetic_2

        z_file_path_1 = AURORA_RESULTS_PATH.joinpath("test2.zss")
        z_file_path_2 = AURORA_RESULTS_PATH.joinpath("test2_from_stored_fc.zss")
        tf1 = process_synthetic_2(force_make_mth5=True, z_file_path=z_file_path_1)
        tf2 = process_synthetic_2(force_make_mth5=False, z_file_path=z_file_path_2)
        assert tf1==tf2


def main():
    # test_case = TestAddFourierCoefficientsToSyntheticData()
    # test_case.test_create_then_use_stored_fcs_for_processing()
    # test_case.setUpClass()
    # test_case.test_123()
    # test_case.test_decimation_and_stft_config_creator()
    # print("se funciona!")
    unittest.main()


if __name__ == "__main__":
    main()



import logging
import pathlib
import unittest

from aurora.pipelines.process_mth5 import process_mth5
from aurora.pipelines.run_summary import RunSummary
from aurora.pipelines.transfer_function_kernel import TransferFunctionKernel
from aurora.test_utils.synthetic.make_mth5_from_asc import create_test1_h5
from aurora.test_utils.synthetic.make_mth5_from_asc import create_test2_h5
from aurora.test_utils.synthetic.make_mth5_from_asc import create_test12rr_h5
from aurora.test_utils.synthetic.make_processing_configs import (
    create_test_run_config,
)
from aurora.test_utils.synthetic.paths import AURORA_RESULTS_PATH
from aurora.transfer_function.kernel_dataset import KernelDataset
from mth5.helpers import close_open_files


# =============================================================================
#  Tests
# =============================================================================


class TestSyntheticProcessing(unittest.TestCase):
    """
    Runs several synthetic processing tests from config creation to tf_cls.

    """

    def setUp(self):
        close_open_files()
        self.file_version = "0.1.0"
        logging.getLogger("matplotlib.font_manager").disabled = True
        logging.getLogger("matplotlib.ticker").disabled = True

    def test_no_crash_with_too_many_decimations(self):
        z_file_path = AURORA_RESULTS_PATH.joinpath("syn1_tfk.zss")
        xml_file_base = "syn1_tfk.xml"
        xml_file_name = AURORA_RESULTS_PATH.joinpath(xml_file_base)
        tf_cls = process_synthetic_1(
            config_keyword="test1_tfk", z_file_path=z_file_path
        )
        tf_cls.write(fn=xml_file_name, file_type="emtfxml")
        tf_cls.write(
            fn=z_file_path.parent.joinpath(f"{z_file_path.stem}_from_tf.zss"),
            file_type="zss",
        )

        xml_file_base = "syn1r2_tfk.xml"
        xml_file_name = AURORA_RESULTS_PATH.joinpath(xml_file_base)
        tf_cls = process_synthetic_1r2(config_keyword="test1r2_tfk")
        tf_cls.write(fn=xml_file_name, file_type="emtfxml")

    def test_can_output_tf_class_and_write_tf_xml(self):
        tf_cls = process_synthetic_1(file_version=self.file_version)
        xml_file_base = "syn1_mth5-010.xml"
        xml_file_name = AURORA_RESULTS_PATH.joinpath(xml_file_base)
        tf_cls.write(fn=xml_file_name, file_type="emtfxml")

    def test_can_use_channel_nomenclature(self):
        channel_nomencalture = "LEMI12"
        z_file_path = AURORA_RESULTS_PATH.joinpath(f"syn1-{channel_nomencalture}.zss")
        tf_cls = process_synthetic_1(
            z_file_path=z_file_path,
            file_version=self.file_version,
            channel_nomenclature=channel_nomencalture,
        )
        xml_file_base = f"syn1_mth5-{self.file_version}_{channel_nomencalture}.xml"
        xml_file_name = AURORA_RESULTS_PATH.joinpath(xml_file_base)
        tf_cls.write(fn=xml_file_name, file_type="emtfxml")

    def test_can_use_mth5_file_version_020(self):
        file_version = "0.2.0"
        z_file_path = AURORA_RESULTS_PATH.joinpath(f"syn1-{file_version}.zss")
        tf_cls = process_synthetic_1(z_file_path=z_file_path, file_version=file_version)
        xml_file_base = f"syn1_mth5v{file_version}.xml"
        xml_file_name = AURORA_RESULTS_PATH.joinpath(xml_file_base)
        tf_cls.write(fn=xml_file_name, file_type="emtfxml")
        tf_cls.write(
            fn=z_file_path.parent.joinpath(f"{z_file_path.stem}_from_tf.zss"),
            file_type="zss",
        )

    def test_can_use_scale_factor_dictionary(self):
        """
        2022-05-13: Added a duplicate run of process_synthetic_1, which is intended to
        test the channel_scale_factors in the new mt_metadata processing class.
        Expected outputs are four .png:

        xy_syn1.png : Shows expected 100 Ohm-m resisitivity
        xy_syn1-scaled.png : Overestimates by 4x for 300 Ohm-m resistivity
        yx_syn1.png : Shows expected 100 Ohm-m resisitivity
        yx_syn1-scaled.png : Underestimates by 4x for 25 Ohm-m resistivity
        These .png are stores in aurora_results folder

        """
        z_file_path = AURORA_RESULTS_PATH.joinpath("syn1-scaled.zss")
        tf_cls = process_synthetic_1(
            z_file_path=z_file_path,
            test_scale_factor=True,
        )
        tf_cls.write(
            fn=z_file_path.parent.joinpath(f"{z_file_path.stem}_from_tf.zss"),
            file_type="zss",
        )

    def test_simultaneous_regression(self):
        z_file_path = AURORA_RESULTS_PATH.joinpath("syn1_simultaneous_estimate.zss")
        tf_cls = process_synthetic_1(
            z_file_path=z_file_path, simultaneous_regression=True
        )
        xml_file_base = "syn1_simultaneous_estimate.xml"
        xml_file_name = AURORA_RESULTS_PATH.joinpath(xml_file_base)
        tf_cls.write(fn=xml_file_name, file_type="emtfxml")
        tf_cls.write(
            fn=z_file_path.parent.joinpath(f"{z_file_path.stem}_from_tf.zss"),
            file_type="zss",
        )

    def test_can_process_other_station(self, force_make_mth5=True):
        tf_cls = process_synthetic_2(force_make_mth5=force_make_mth5)
        xml_file_name = AURORA_RESULTS_PATH.joinpath("syn2.xml")
        tf_cls.write(fn=xml_file_name, file_type="emtfxml")

    def test_can_process_remote_reference_data(self):
        tf_cls = process_synthetic_1r2(channel_nomenclature="default")
        xml_file_base = "syn12rr_mth5-010.xml"
        xml_file_name = AURORA_RESULTS_PATH.joinpath(xml_file_base)
        tf_cls.write(
            fn=xml_file_name,
            file_type="emtfxml",
        )

    def test_can_process_remote_reference_data_with_channel_nomenclature(self):
        tf_cls = process_synthetic_1r2(channel_nomenclature="LEMI34")
        xml_file_base = "syn12rr_mth5-010_LEMI34.xml"
        xml_file_name = AURORA_RESULTS_PATH.joinpath(xml_file_base)
        tf_cls.write(
            fn=xml_file_name,
            file_type="emtfxml",
        )


def process_synthetic_1(
    config_keyword="test1",
    z_file_path="",
    test_scale_factor=False,
    simultaneous_regression=False,
    file_version="0.1.0",
    return_collection=False,
    channel_nomenclature="default",
    reload_config=False,
):
    """

    Parameters
    ----------
    config_keyword: str
        "test1", "test1_tfk", this is an argument passed to the create_test_run_config
        as test_case_id.
    z_file_path: str or path
        Where the z-file will be output
    test_scale_factor: bool
        If true, will assign scale factors to the channels
    simultaneous_regression: bool
        If True will do regression all outut channels in one step, rather than the
        usual, channel-by-channel method
    file_version: str
        one of ["0.1.0", "0.2.0"]
    Returns
    -------
    tf_result: TransferFunctionCollection or mt_metadata.transfer_functions.TF
        Should change so that it is mt_metadata.TF (see Issue #143)
    """
    mth5_path = create_test1_h5(
        file_version=file_version, channel_nomenclature=channel_nomenclature
    )
    run_summary = RunSummary()
    run_summary.from_mth5s(
        [
            mth5_path,
        ]
    )
    # next two lines purely for codecov
    run_summary.print_mini_summary
    run_summary_clone = run_summary.clone()
    # run_summary.drop_runs_shorter_than(100000)
    tfk_dataset = KernelDataset()
    tfk_dataset.from_run_summary(run_summary_clone, "test1")

    # Test that channel_scale_factors column is optional
    if test_scale_factor:
        scale_factors = {
            "ex": 10.0,
            "ey": 3.0,
            "hx": 6.0,
            "hy": 5.0,
            "hz": 100.0,
        }
        tfk_dataset.df["channel_scale_factors"].at[0] = scale_factors
    else:
        tfk_dataset.df.drop(columns=["channel_scale_factors"], inplace=True)

    processing_config = create_test_run_config(
        config_keyword, tfk_dataset, channel_nomenclature=channel_nomenclature
    )
    # Relates to issue #172
    # reload_config = True
    # if reload_config:
    #     from mt_metadata.transfer_functions.processing.aurora import Processing
    #     p = Processing()
    #     config_path = pathlib.Path("config")
    #     json_fn = config_path.joinpath(processing_config.json_fn())
    #     p.from_json(json_fn)

    if simultaneous_regression:
        for decimation in processing_config.decimations:
            decimation.estimator.estimate_per_channel = False

    tf_result = process_mth5(
        processing_config,
        tfk_dataset=tfk_dataset,
        z_file_path=z_file_path,
        return_collection=return_collection,
    )

    if return_collection:
        z_figure_name = z_file_path.name.replace("zss", "png")
        for xy_or_yx in ["xy", "yx"]:
            ttl_str = f"{xy_or_yx} component, test_scale_factor = {test_scale_factor}"
            out_png_name = f"{xy_or_yx}_{z_figure_name}"
            tf_result.rho_phi_plot(
                xy_or_yx=xy_or_yx,
                ttl_str=ttl_str,
                show=False,
                figure_basename=out_png_name,
                figure_path=AURORA_RESULTS_PATH,
            )
    return tf_result


def process_synthetic_2(force_make_mth5=True, z_file_path=None):
    """"""
    station_id = "test2"
    mth5_path = create_test2_h5(force_make_mth5=force_make_mth5)
    mth5_paths = [
        mth5_path,
    ]
    run_summary = RunSummary()
    run_summary.from_mth5s(mth5_paths)
    tfk_dataset = KernelDataset()
    tfk_dataset.from_run_summary(run_summary, station_id)
    processing_config = create_test_run_config(station_id, tfk_dataset)
    for decimation_level in processing_config.decimations:
        decimation_level.save_fcs = True
        decimation_level.save_fcs_type = "h5"
        # decimation_level.save_fcs_type = "csv"
    tfc = process_mth5(
        processing_config,
        tfk_dataset=tfk_dataset,
        z_file_path=z_file_path,
    )
    return tfc


def process_synthetic_1r2(
    config_keyword="test1r2",
    channel_nomenclature="default",
    return_collection=False,
):
    mth5_path = create_test12rr_h5(channel_nomenclature=channel_nomenclature)
    mth5_paths = [
        mth5_path,
    ]
    run_summary = RunSummary()
    run_summary.from_mth5s(mth5_paths)
    tfk_dataset = KernelDataset()
    tfk_dataset.from_run_summary(run_summary, "test1", "test2")
    processing_config = create_test_run_config(
        config_keyword, tfk_dataset, channel_nomenclature=channel_nomenclature
    )
    tfc = process_mth5(
        processing_config,
        tfk_dataset=tfk_dataset,
        return_collection=return_collection,
    )
    return tfc


def main():
    """
    Testing the processing of synthetic data
    """
    # tmp = TestSyntheticProcessing()
    # tmp.setUp()
    # tmp.test_can_process_other_station() # makes FC csvs

    # tmp.test_can_output_tf_class_and_write_tf_xml()
    # tmp.test_no_crash_with_too_many_decimations()
    # tmp.test_can_use_scale_factor_dictionary()

    unittest.main()


if __name__ == "__main__":
    main()


# def process_synthetic_1_underdetermined():
#     """
#     Just like process_synthetic_1, but the window is ridiculously long so that we
#     encounter the underdetermined problem. We actually pass that test but in testing
#     I found that at the next band over, which has more data because there are multipe
#     FCs the sigma in TRME comes out as negative. see issue #4 and issue #55.
#     Returns
#     -------
#
#     """
#     test_config = CONFIG_PATH.joinpath("test1_run_config_underdetermined.json")
#     # test_config = Path("config", "test1_run_config_underdetermined.json")
#     run_id = "001"
#     process_mth5(test_config, run_id, units="MT")
#
#
# def process_synthetic_1_with_nans():
#     """
#
#     Returns
#     -------
#
#     """
#     test_config = CONFIG_PATH.joinpath("test1_run_config_nan.json")
#     #    test_config = Path("config", "test1_run_config_nan.json")
#     run_id = "001"
#     process_mth5(test_config, run_id, units="MT")



from aurora.sandbox.io_helpers.garys_matlab_zfiles.matlab_z_file_reader import (
    test_matlab_zfile_reader,
)


def test():
    test_matlab_zfile_reader(case_id="IAK34ss")
    # test_matlab_zfile_reader(case_id="synthetic")


if __name__ == "__main__":
    test()



import numpy as np
import pandas as pd
import time


def fast_arange(t0, n_samples, sample_rate):
    t0 = np.datetime64(t0)
    dt = 1.0 / sample_rate
    dt_nanoseconds = int(np.round(1e9 * dt))
    dt_timedelta = np.timedelta64(dt_nanoseconds, "ns")
    time_index = t0 + np.arange(n_samples) * dt_timedelta
    return time_index


def slow_comprehension(t0, n_samples, sample_rate):
    t0 = np.datetime64(t0)
    dt = 1.0 / sample_rate
    time_vector_seconds = dt * np.arange(n_samples)
    time_vector_nanoseconds = (np.round(1e9 * time_vector_seconds)).astype(int)
    time_index = np.array(
        [t0 + np.timedelta64(x, "ns") for x in time_vector_nanoseconds]
    )
    return time_index


TIME_AXIS_GENERATOR_FUNCTIONS = {}
TIME_AXIS_GENERATOR_FUNCTIONS["fast_arange"] = fast_arange
TIME_AXIS_GENERATOR_FUNCTIONS["slow_comprehension"] = slow_comprehension


def decide_time_axis_method(sample_rate):
    dt = 1.0 / sample_rate
    ns_per_sample = 1e9 * dt
    if np.floor(ns_per_sample) == np.ceil(ns_per_sample):
        method = "fast_arange"
    else:
        method = "slow_comprehension"
    return method


def make_time_axis(t0, n_samples, sample_rate):
    method = decide_time_axis_method(sample_rate)
    time_axis = TIME_AXIS_GENERATOR_FUNCTIONS[method](t0, n_samples, sample_rate)
    return time_axis


def test_generate_time_axis(t0, n_samples, sample_rate):
    """Two obvious ways to generate an axis of timestamps here. One method is slow and
    more precise, the other is fast but drops some nanoseconds due to integer
    roundoff error.

    To see this, consider the example of say 3Hz, we are 333333333ns between samples,
    which drops 1ns per second if we scale a nanoseconds=np.arange(N)
    The issue here is that the nanoseconds granularity forces a roundoff error


    Probably will use logic like:
    if there_are_integer_ns_per_sample:
        time_stamps = do_it_the_fast_way()
    else:
        time_stamps = do_it_the_slow_way()
    return time_stamps

    Parameters
    ----------
    t0
    n_samples
    sample_rate

    Returns
    -------

    """
    t0 = np.datetime64(t0)

    # SLOW
    tt = time.time()
    time_index_1 = slow_comprehension(t0, n_samples, sample_rate)
    processing_time_1 = tt - time.time()
    print(f"processing_time_1 = {processing_time_1}")
    

    # FAST
    tt = time.time()
    time_index_2 = fast_arange(t0, n_samples, sample_rate)
    processing_time_2 = tt - time.time()
    print(f"processing_time_2 {processing_time_2}")
    print(f"ratio of processing times {processing_time_1/processing_time_2}")
    
    if (np.abs(time_index_2 - time_index_1)).sum() == 0:
        pass
    else:
        print("Time axes are not equal")
    return time_index_1


# def make_time_axis_method1(t0, n_samples, sample_rate):
# pass


def do_some_tests():
    n_samples = 1000
    sample_rate = 50.0  # Hz
    t0 = pd.Timestamp(1977, 3, 2, 6, 1, 44)
    time_axis = test_generate_time_axis(t0, n_samples, sample_rate)
    print(f"{time_axis[0]} ...{time_axis[-1]}")
    sample_rate = 3.0  # Hz
    time_axis = test_generate_time_axis(t0, n_samples, sample_rate)


def main():
    do_some_tests()


if __name__ == "__main__":
    main()



from mt_metadata.timeseries.filters.coefficient_filter import CoefficientFilter
from mt_metadata.timeseries.filters.frequency_response_table_filter import (
    FrequencyResponseTableFilter,
)
from aurora.general_helper_functions import TEST_PATH


def make_coefficient_filter(gain=1.0, name="generic coefficient filter", **kwargs):
    """

    Parameters
    ----------
    gain
    name
    units_in : string
        one of "digital counts", "millivolts", etc.
        TODO: Add a refernce here to the list of units supported in mt_metadata

    Returns
    -------

    """
    # in general, you need to add all required fields from the standards.json
    default_units_in = "unknown"
    default_units_out = "unknown"

    cf = CoefficientFilter()
    cf.gain = gain
    cf.name = name

    cf.units_in = kwargs.get("units_in", default_units_in)
    cf.units_out = kwargs.get("units_out", default_units_out)

    return cf


def make_frequency_response_table_filter(case="bf4"):
    fap_filter = FrequencyResponseTableFilter()
    if case == "bf4":
        import numpy as np
        import pandas as pd

        bf4_path = TEST_PATH.joinpath("parkfield", "bf4_9819.csv")
        df = pd.read_csv(bf4_path)  # , skiprows=1)
        # Hz, V/nT, degrees
        fap_filter.frequencies = df["Frequency [Hz]"].values
        fap_filter.amplitudes = df["Amplitude [V/nT]"].values
        fap_filter.phases = np.deg2rad(df["Phase [degrees]"].values)
        fap_filter.units_in = "volts"
        fap_filter.units_out = "nanotesla"
        fap_filter.gain = 1.0
        fap_filter.name = "bf4"
        return fap_filter
    else:
        print(f"case {case} not supported for FAP Table")
        raise Exception


def make_volt_per_meter_to_millivolt_per_km_converter():
    """
    This represents a filter that HAS ALREADY converted from mV/km to V/m.
    This means it is as if the data were multiplied by 1e-6.  To correct for this
    in the data we will want to multiply by 1e6 which is what will happen when we
    divide by the filter gain.
    Returns
    -------

    """
    coeff_filter = make_coefficient_filter(
        gain=1e-6,
        units_in="millivolts per kilometer",
        units_out="volts per meter",
        name="MT to SI electric field conversion",
    )
    return coeff_filter


MT2SI_ELECTRIC_FIELD_FILTER = make_volt_per_meter_to_millivolt_per_km_converter()


def main():
    make_volt_per_meter_to_millivolt_per_km_converter()



import numpy as np


def rme_beta(r0):
    """
    This is an RME specific property.
    It represents a bias in the calculation of residual_variance
    which we correct for in TRME and TRME_RR.

    The implemented formula is an approximation.  This is approximately equal to 1/beta
    where beta is defined by Equation A3 in Egbert & Booker 1986.

    for r0=1.5  the correction factor is ~1.28 and beta ~0.78

    Some notes from a discussion with Gary:
    In the regression esimtate you downweight things with large errors, but
    you need to define what's large.  You estimate the standard devation
    (sigma) of the errors from the residuals BUT with this cleaned data
    approach (Yc) sigma is smaller than it should be, you need to
    compensate for this by using a correction_factor. It's basically the
    expectation, if the data really were Gaussian, and you estimated from
    the corrected data. This is how much too small the estimate would be.

    If you change the penalty functional you may need a pencil, paper and
    some calculus.  The relationship between the corrected-data-residuals
    and the gaussin residauls could change if you change the penalty

    Parameters
    ----------
    r0 : float
        The number of standard errors at which the RME method transitions from
        quadratic weighting to linear

    Returns
    -------
    beta : float
        correction factor = 1/beta
    """
    beta = 1.0 - np.exp(-r0)
    return beta



from pathlib import Path

from mt_metadata.transfer_functions.processing.aurora import Processing


def initialize_config(processing_config):
    """

    Parameters
    ----------
    processing_cfg: path or str
        Either an instance of the processing class or a path to a json file that a
        Processing object is stored in.
    Returns :
    -------
    config: mt_metadata.transfer_functions.processing.aurora.Processing
        Object that contains the processing parameters
    """
    if isinstance(processing_config, (Path, str)):
        config = Processing()
        config.from_json(processing_config)
    elif isinstance(processing_config, Processing):
        config = processing_config
    else:
        raise Exception(f"Unrecognized config of type {type(processing_config)}")
    return config



def make_subtitle(
    rho_rms_aurora,
    rho_rms_emtf,
    phi_rms_aurora,
    phi_rms_emtf,
    matlab_or_fortran,
    ttl_str="",
):
    """

    Parameters
    ----------
    rho_rms_aurora: float
        rho_rms for aurora data differenced against a model. comes from compute_rms
    rho_rms_emtf:
        rho_rms for emtf data differenced against a model. comes from compute_rms
    phi_rms_aurora:
        phi_rms for aurora data differenced against a model. comes from compute_rms
    phi_rms_emtf:
        phi_rms for emtf data differenced against a model. comes from compute_rms
    matlab_or_fortran: str
        "matlab" or "fortran".  A specifer for the version of emtf.
    ttl_str: str
        string onto which we add the subtitle

    Returns
    -------
    ttl_str: str
        Figure title with subtitle

    """
    ttl_str += (
        f"\n rho rms_aurora {rho_rms_aurora:.1f} rms_{matlab_or_fortran}"
        f" {rho_rms_emtf:.1f}"
    )
    ttl_str += (
        f"\n phi rms_aurora {phi_rms_aurora:.1f} rms_{matlab_or_fortran}"
        f" {phi_rms_emtf:.1f}"
    )
    return ttl_str


def make_figure_basename(
    local_station_id, remote_station_id, xy_or_yx, matlab_or_fortran
):
    """

    Parameters
    ----------
    local_station_id: str
        station label
    remote_station_id: str
        remote reference station label
    xy_or_yx: str
        mode: "xy" or "yx"
    matlab_or_fortran: str
        "matlab" or "fortran".  A specifer for the version of emtf.

    Returns
    -------
    figure_basename: str
        filename for figure

    """
    station_string = f"{local_station_id}"
    if remote_station_id:
        station_string = f"{station_string}_rr{remote_station_id}"
    figure_basename = f"synthetic_{station_string}_{xy_or_yx}_{matlab_or_fortran}.png"
    return figure_basename


def plot_rho_phi(
    xy_or_yx,
    tf_collection,
    rho_rms_aurora,
    rho_rms_emtf,
    phi_rms_aurora,
    phi_rms_emtf,
    matlab_or_fortran,
    aux_data=None,
    use_subtitle=True,
    show_plot=False,
    output_path=None,
):
    """
    Could be made into a method of TF Collection
    Parameters
    ----------
    xy_or_yx
    tf_collection
    rho_rms_aurora
    rho_rms_emtf
    phi_rms_aurora
    phi_rms_emtf
    matlab_or_fortran
    aux_data
    use_subtitle
    show_plot

    Returns
    -------

    """
    ttl_str = ""
    if use_subtitle:
        ttl_str = make_subtitle(
            rho_rms_aurora,
            rho_rms_emtf,
            phi_rms_aurora,
            phi_rms_emtf,
            matlab_or_fortran,
        )

    figure_basename = make_figure_basename(
        tf_collection.local_station_id,
        tf_collection.remote_station_id,
        xy_or_yx,
        matlab_or_fortran,
    )
    tf_collection.rho_phi_plot(
        aux_data=aux_data,
        xy_or_yx=xy_or_yx,
        ttl_str=ttl_str,
        show=show_plot,
        figure_basename=figure_basename,
        figure_path=output_path,
    )
    return



"""
This is a container to hold:
1. TransferFunctionHeader
2. Dictionary of TransferFunction Objects

Note that a single transfer function object is associated with a station,
which we call the "local_station".  In a database of TFs we could add a column
for local_station and one for reference station.
"""

import numpy as np
import xarray as xr

from aurora.transfer_function.plot.rho_phi_helpers import plot_phi
from aurora.transfer_function.plot.rho_phi_helpers import plot_rho

EMTF_REGRESSION_ENGINE_LABELS = {}
EMTF_REGRESSION_ENGINE_LABELS["RME"] = "Robust Single Station"
EMTF_REGRESSION_ENGINE_LABELS["RME_RR"] = "Robust Remote Reference"


class TransferFunctionCollection(object):
    def __init__(self, **kwargs):
        """

        Parameters
        ----------
        kwargs

        tf_dict: dict
            This is a dictionary of TTFZ objects, one per decimation level.  They are
            keyed by the decimation_level_id, usually integers 0, 1, 2... n_dec

        """
        self.tf_dict = kwargs.get("tf_dict", None)
        self.processing_config = kwargs.get("processing_config", None)
        self.labelled_tf = None
        self.merged_tf = None
        self.merged_cov_nn = None
        self.merged_cov_ss_inv = None

    @property
    def header(self):
        return self.tf_dict[0].tf_header

    @property
    def local_station_id(self):
        return self.header.local_station.id

    @property
    def remote_station_id(self):
        if self.header.remote_station:
            return self.header.remote_station[0].id
        else:
            return ""

    @property
    def total_number_of_frequencies(self):
        num_frequecies = 0
        for dec_level in self.tf_dict.keys():
            num_frequecies += len(self.tf_dict[dec_level].periods)
        return num_frequecies

    @property
    def channel_list(self):
        tf_input_chs = self.tf_dict[0].transfer_function.input_channel.data.tolist()
        tf_output_chs = self.tf_dict[0].transfer_function.output_channel.data.tolist()
        all_channels = tf_input_chs + tf_output_chs
        return all_channels

    @property
    def total_number_of_channels(self):
        num_channels = 0
        num_channels += len(self.header.input_channels)
        num_channels += len(self.header.output_channels)
        return num_channels

    @property
    def number_of_decimation_levels(self):
        return len(self.tf_dict)

    def get_merged_dict(self, channel_nomenclature):
        output = {}
        self._merge_decimation_levels()
        self.check_all_channels_present(channel_nomenclature)
        output["tf"] = self.merged_tf
        output["tf_xarray"] = self.labelled_tf
        output["cov_ss_inv"] = self.merged_cov_ss_inv
        output["cov_nn"] = self.merged_cov_nn
        return output

    def _merge_decimation_levels(self):
        """
        Will merge all decimation levels into a single xarray for export.
        The output of this may become its own class, MergedTransferFunction

        One concern here is that the same period can be estiamted at more then one
        decimation level, making the frequency or period axis not of the same order
        as the number of estimates.  Ways around this:
         1. We can multi-index the period axis with decimation level
         2. Make sure that the processing config does not dupliacate periods
         3. Take the average estimate over all periods heving the same value
         4. Drop all but one estimate accoring to a rule (keep most observations say)

        Does TFXML support multiple estimates? How is it dealt with there?
        There is an issue here -

        Flow:
        starting from self.tf_dict, we cast each decimation level's tf, which is
        nativley an xr.DataArray to xr.Dataset, forming a list of datasets.
        This dataset list is then combined over all periods forming a representation
        of the TTFZ which is merged over all the decimation levels.

        result is to build:
        merged_tf
        merged_cov_ss_inv
        merged_cov_nn
        """
        n_dec = self.number_of_decimation_levels

        tmp = [self.tf_dict[i].tf.to_dataset("period") for i in range(n_dec)]
        merged_tf = xr.combine_by_coords(tmp)
        merged_tf = merged_tf.to_array(dim="period")
        self.merged_tf = merged_tf

        tmp = [self.tf_dict[i].cov_ss_inv.to_dataset("period") for i in range(n_dec)]
        merged_cov_ss_inv = xr.combine_by_coords(tmp)
        merged_cov_ss_inv = merged_cov_ss_inv.to_array("period")
        self.merged_cov_ss_inv = merged_cov_ss_inv

        tmp = [self.tf_dict[i].cov_nn.to_dataset("period") for i in range(n_dec)]
        merged_cov_nn = xr.combine_by_coords(tmp)
        merged_cov_nn = merged_cov_nn.to_array("period")
        self.merged_cov_nn = merged_cov_nn

        return

    def check_all_channels_present(self, channel_nomenclature):
        """

        Parameters
        ----------
        channel_nomenclature: mt_metadata.transfer_functions.processing.aurora.channel_nomenclature.ChannelNomenclature
            Scheme according to how channels are named
        """
        ex, ey, hx, hy, hz = channel_nomenclature.unpack()
        if hz not in self.merged_tf.output_channel.data.tolist():
            output_channels_original = self.merged_tf.output_channel.data.tolist()
            output_channels = [f"tmp___{x}" for x in output_channels_original]
            output_channels[0] = hz
            tmp = self.merged_tf.copy(deep=True)
            tmp = tmp.assign_coords({"output_channel": output_channels})
            tmp.data *= np.nan
            tmp = tmp.to_dataset("period")
            tmp = tmp.merge(self.merged_tf.to_dataset("period"))
            tmp = tmp.to_array("period")
            output_channels = tmp.output_channel.data.tolist()
            output_channels = [x for x in output_channels if x[0:6] != "tmp___"]
            tmp = tmp.sel(output_channel=output_channels)
            self.merged_tf = tmp

            n_output_ch = len(self.merged_tf.output_channel)  # 3
            n_periods = len(self.merged_tf.period)
            cov_nn_dims = (n_output_ch, n_output_ch, n_periods)
            cov_nn_data = np.zeros(cov_nn_dims, dtype=np.complex128)
            import xarray as xr

            cov_nn = xr.DataArray(
                cov_nn_data,
                dims=["output_channel_1", "output_channel_2", "period"],
                coords={
                    "output_channel_1": self.merged_tf.output_channel.data,
                    "output_channel_2": self.merged_tf.output_channel.data,
                    "period": self.merged_tf.period,
                },
            )

            # to dataset and back makes the coords line up -- this does not seem robust
            cov_nn = cov_nn.to_dataset("period").to_array("period")

            for out_ch_1 in cov_nn.output_channel_1.data.tolist():
                for out_ch_2 in cov_nn.output_channel_1.data.tolist():
                    try:
                        values = self.merged_cov_nn.loc[:, out_ch_1, out_ch_2]
                        cov_nn.loc[:, out_ch_1, out_ch_2] = values
                    except KeyError:
                        pass
            self.merged_cov_nn = cov_nn

    def rho_phi_plot(
        self,
        show=True,
        aux_data=None,
        xy_or_yx="xy",
        ttl_str="",
        x_axis_fontsize=25,
        y_axis_fontsize=25,
        ttl_fontsize=16,
        markersize=10,
        rho_ylims=[10, 1000],
        phi_ylims=[0, 90],
        **kwargs,
    ):
        """
        One-off plotting method intended only for the synthetic test data for aurora dev
        Parameters
        ----------
        show
        aux_data
        xy_or_yx
        ttl_str

        Returns
        -------

        """

        import matplotlib.pyplot as plt

        fig, axs = plt.subplots(nrows=2, figsize=(11, 8.5), dpi=300)
        # plotter.rho_sub_plot(axs[0], ttl_str=ttl_str)
        # plotter.phase_sub_plot(axs[1], ttl_str=ttl_str)

        color_cyc = {0: "r", 1: "orange", 2: "g", 3: "b", 4: "k"}
        for i_dec in self.tf_dict.keys():
            tf = self.tf_dict[i_dec]

            if xy_or_yx == "xy":
                aurora_rho = tf.rho[:, 0]
            else:
                aurora_rho = tf.rho[:, 1]
            plot_rho(
                axs[0],
                tf.periods,
                aurora_rho,
                marker="o",
                color=color_cyc[i_dec],
                linestyle="None",
                label=f"aurora {i_dec}",
                markersize=markersize,
            )

            if xy_or_yx == "xy":
                aurora_phi = tf.phi[:, 0]
            else:
                aurora_phi = tf.phi[:, 1]
            # rotate phases so all are positive:
            negative_phi_indices = np.where(aurora_phi < 0)[0]
            aurora_phi[negative_phi_indices] += 180.0
            plot_phi(
                axs[1],
                tf.periods,
                aurora_phi,
                marker="o",
                color=color_cyc[i_dec],
                linestyle="None",
                label=f"aurora {i_dec}",
                markersize=markersize,
            )

        if aux_data:
            #            try:
            decimation_levels = list(set(aux_data.decimation_levels))
            shape_cyc = {0: "s", 1: "v", 2: "*", 3: "^"}
            if xy_or_yx == "xy":
                emtf_rho = aux_data.rxy
                emtf_phi = aux_data.pxy
            elif xy_or_yx == "yx":
                emtf_rho = aux_data.ryx
                emtf_phi = aux_data.pyx

            axs[0].loglog(axs[0].get_xlim(), 100 * np.ones(2), color="k")
            axs[1].semilogx(axs[1].get_xlim(), 45 * np.ones(2), color="k")
            for i_dec in decimation_levels:

                ndx = np.where(aux_data.decimation_levels == i_dec)[0]
                axs[0].loglog(
                    aux_data.periods[ndx],
                    emtf_rho[ndx],
                    marker=shape_cyc[i_dec - 1],
                    color="k",
                    linestyle="None",
                    label=f"emtf " f"{int(i_dec-1)}",
                    markersize=markersize,
                )
                axs[1].semilogx(
                    aux_data.periods[ndx],
                    emtf_phi[ndx],
                    marker=shape_cyc[i_dec - 1],
                    color="k",
                    linestyle="None",
                    markersize=markersize,
                    label=f"emtf " f"{int(i_dec-1)}",
                )

        axs[0].legend(ncol=2)
        axs[1].legend(ncol=2)

        axs[1].set_xlabel("Period (s)", fontsize=x_axis_fontsize)
        axs[0].set_ylabel(r"$\Omega$-m", fontsize=y_axis_fontsize)
        axs[1].set_ylabel("Degrees", fontsize=y_axis_fontsize)

        ttl_str = f"{tf.tf_header.local_station.id} {xy_or_yx} \n{ttl_str}"
        axs[0].set_title(ttl_str, fontsize=ttl_fontsize)
        if rho_ylims is not None:
            axs[0].set_ylim(rho_ylims)
        if phi_ylims is not None:
            axs[1].set_ylim(phi_ylims)

        from aurora.general_helper_functions import FIGURES_PATH

        default_figure_basename = f"{self.local_station_id}_{xy_or_yx}.png"
        figure_basename = kwargs.get("figure_basename", default_figure_basename)
        figure_path = kwargs.get("figure_path", FIGURES_PATH)
        out_file = figure_path.joinpath(figure_basename)
        plt.savefig(out_file)
        if show:
            plt.show()
        plt.close(fig)



"""

Note 1: Functionality of RunSummary()
1. User can get a list of local_station options, which correspond to unique pairs
of values: (survey_id,  station_id)

2. User can see all possible ways of processing the data:
- one list per (survey_id,  station_id) pair in the run_summary

Some of the following functionalities may end up in KernelDataset:
3. User can select local_station
-this can trigger a reduction of runs to only those that are from the local staion
and simultaneous runs at other stations
4. Given a local station, a list of possible reference stations can be generated
5. Given a remote reference station, a list of all relevent runs, truncated to
maximize coverage of the local station runs is generated
6. Given such a "restricted run list", runs can be dropped
7. Time interval endpoints can be changed


"""

import copy
import pandas as pd

import mth5
from mth5.utils.helpers import initialize_mth5


INPUT_CHANNELS = [
    "hx",
    "hy",
]
OUTPUT_CHANNELS = [
    "ex",
    "ey",
    "hz",
]
RUN_SUMMARY_COLUMNS = [
    "survey",
    "station_id",
    "run_id",
    "start",
    "end",
    "sample_rate",
    "input_channels",
    "output_channels",
    "remote",
    "mth5_path",
]


class RunSummary:
    """
    The dependencies aren't clear yet.
    Maybe still Dataset:
        Could have methods
            "drop_runs_shorter_than"
            "fill_gaps_by_time_interval"
            "fill_gaps_by_run_names"
            "

    For the full MMT case this may need modification to a channel based summary.

    Question: To return a copy or modify in-place when querying.  Need to decide on
    standards and syntax.  Handling this in general could use a decorator that allows
    df as kwarg, and if it is not passed the modification is done in place.
    The user who doesn't want to modify in place can work with a clone.
    Could also try the @staticmethod decorator so that it returns a modified df.

    """

    def __init__(self, **kwargs):
        self.column_dtypes = [str, str, pd.Timestamp, pd.Timestamp]
        self._input_dict = kwargs.get("input_dict", None)
        self.df = kwargs.get("df", None)
        self._mini_summary_columns = ["survey", "station_id", "run_id", "start", "end"]

    def clone(self):
        """
        2022-10-20:
        Cloning may be causing issues with extra instances of open h5 files ...

        """
        return copy.deepcopy(self)

    def from_mth5s(self, mth5_list):
        run_summary_df = extract_run_summaries_from_mth5s(mth5_list)
        self.df = run_summary_df

    @property
    def mini_summary(self):
        return self.df[self._mini_summary_columns]

    @property
    def print_mini_summary(self):
        print(self.mini_summary)

    def add_duration(self, df=None):
        """

        Parameters
        ----------
        df

        """
        if df is None:
            df = self.df
        timedeltas = df.end - df.start
        durations = [x.total_seconds() for x in timedeltas]
        df["duration"] = durations
        return

    def check_runs_are_valid(self, drop=False, **kwargs):
        """kwargs can tell us what sorts of conditions to check, for example all_zero, there are nan, etc."""
        # check_for_all_zero_runs
        for i_row, row in self.df.iterrows():
            print(f"Checking row for zeros {row}")
            m = mth5.mth5.MTH5()
            m.open_mth5(row.mth5_path)
            run_obj = m.get_run(row.station_id, row.run_id, row.survey)
            runts = run_obj.to_runts()
            if runts.dataset.to_array().data.__abs__().sum() == 0:
                print("CRITICAL: Detected a run with all zero values")
                self.df["valid"].at[i_row] = False
            # load each run, and take the median of the sum of the absolute values
        if drop:
            self.drop_invalid_rows()
        return

    def drop_invalid_rows(self):
        self.df = self.df[self.df.valid]
        self.df.reset_index(drop=True, inplace=True)

    # BELOW FUNCTION CAN BE COPIED FROM METHOD IN KernelDataset()
    # def drop_runs_shorter_than(self, duration, units="s"):
    #     if units != "s":
    #         raise NotImplementedError
    #     if "duration" not in self.df.columns:
    #         self.add_duration()
    #     drop_cond = self.df.duration < duration
    #     # df = self.df[drop_cond]
    #     self.df.drop(self.df[drop_cond].index, inplace=True)
    #     df = df.reset_index()
    #
    #     self.df = df
    #     return df


def channel_summary_to_run_summary(
    ch_summary,
    allowed_input_channels=INPUT_CHANNELS,
    allowed_output_channels=OUTPUT_CHANNELS,
    sortby=["station_id", "start"],
):
    """
    TODO: replace station_id with station, and run_id with run
    Note will need to modify: aurora/tests/config$ more test_dataset_dataframe.py
    TODO: Add logic for handling input and output channels based on channel
    summary.  Specifically, consider the case where there is no vertical magnetic
    field, this information is available via ch_summary, and output channels should
    then not include hz.
    TODO: Just inherit all the run-level and higher el'ts of the channel_summary,
    including n_samples?

    When creating the dataset dataframe, make it have these columns:
    [
            "station_id",
            "run_id",
            "start",
            "end",
            "mth5_path",
            "sample_rate",
            "input_channels",
            "output_channels",
            "remote",
            "channel_scale_factors",
        ]

    Parameters
    ----------
    ch_summary: mth5.tables.channel_table.ChannelSummaryTable or pandas DataFrame
       If its a dataframe it is a representation of an mth5 channel_summary.
        Maybe restricted to only have certain stations and runs before being passed to
        this method
    allowed_input_channels: list of strings
        Normally ["hx", "hy", ]
        These are the allowable input channel names for the processing.  See further
        note under allowed_output_channels.
    allowed_output_channels: list of strings
        Normally ["ex", "ey", "hz", ]
        These are the allowable output channel names for the processing.
        A global list of these is kept at the top of this module.  The purpose of
        this is to distinguish between runs that have different layouts, for example
        some runs will have hz and some will not, and we cannot process for hz the
        runs that do not have it.  By making this a kwarg we sort of prop the door
        open for more general names (see issue #74).
    sortby: bool or list
        Default: ["station_id", "start"]

    Returns
    -------
    run_summary_df: pd.Dataframe
        A table with one row per "acquistion run" that was in the input channel
        summary table
    """
    if isinstance(ch_summary, mth5.tables.channel_table.ChannelSummaryTable):
        ch_summary_df = ch_summary.to_dataframe()
    elif isinstance(ch_summary, pd.DataFrame):
        ch_summary_df = ch_summary
    group_by_columns = ["survey", "station", "run"]
    grouper = ch_summary_df.groupby(group_by_columns)
    n_station_runs = len(grouper)
    survey_ids = n_station_runs * [None]
    station_ids = n_station_runs * [None]
    run_ids = n_station_runs * [None]
    start_times = n_station_runs * [None]
    end_times = n_station_runs * [None]
    sample_rates = n_station_runs * [None]
    input_channels = n_station_runs * [None]
    output_channels = n_station_runs * [None]
    channel_scale_factors = n_station_runs * [None]
    i = 0
    for group_values, group in grouper:
        group_info = dict(zip(group_by_columns, group_values))  # handy for debug
        # for k, v in group_info.items():
        #     print(f"{k} = {v}")
        survey_ids[i] = group_info["survey"]
        station_ids[i] = group_info["station"]
        run_ids[i] = group_info["run"]
        start_times[i] = group.start.iloc[0]
        end_times[i] = group.end.iloc[0]
        sample_rates[i] = group.sample_rate.iloc[0]
        channels_list = group.component.to_list()
        num_channels = len(channels_list)
        input_channels[i] = [x for x in channels_list if x in allowed_input_channels]
        output_channels[i] = [x for x in channels_list if x in allowed_output_channels]
        channel_scale_factors[i] = dict(zip(channels_list, num_channels * [1.0]))
        i += 1

    data_dict = {}
    data_dict["survey"] = survey_ids
    data_dict["station_id"] = station_ids
    data_dict["run_id"] = run_ids
    data_dict["start"] = start_times
    data_dict["end"] = end_times
    data_dict["sample_rate"] = sample_rates
    data_dict["input_channels"] = input_channels
    data_dict["output_channels"] = output_channels
    data_dict["channel_scale_factors"] = channel_scale_factors
    data_dict["valid"] = True

    run_summary_df = pd.DataFrame(data=data_dict)
    if sortby:
        run_summary_df.sort_values(by=sortby, inplace=True)
    return run_summary_df


def extract_run_summary_from_mth5(mth5_obj, summary_type="run"):
    """

    Parameters
    ----------
    mth5_obj: mth5.mth5.MTH5
        The initialized mth5 object that will be interrogated
    summary_type: str
        One of ["run", "channel"].  Returns a run summary or a channel summary

    Returns
    -------
    out_df: pd.Dataframe
        Table summarizing the available runs in the input mth5_obj
    """
    channel_summary_df = mth5_obj.channel_summary.to_dataframe()
    # check that the mth5 has been summarized already
    if len(channel_summary_df) < 2:
        print("Channel summary maybe not initialized yet, 3 or more channels expected.")
        mth5_obj.channel_summary.summarize()
        channel_summary_df = mth5_obj.channel_summary.to_dataframe()
    if summary_type == "run":
        out_df = channel_summary_to_run_summary(channel_summary_df)
    else:
        out_df = channel_summary_df
    out_df["mth5_path"] = str(mth5_obj.filename)
    return out_df


def extract_run_summaries_from_mth5s(mth5_list, summary_type="run", deduplicate=True):
    """
    ToDo: Move this method into mth5? or mth5_helpers?
    ToDo: Make this a class so that the __repr__ is a nice visual representation of the
    df, like what channel summary does in mth5

    2022-05-28 Modified to allow this method to accept mth5 objects as well as the
    already supported types of pathlib.Path or str

    Given a list of mth5s, this returns a dataframe of all available runs

    In order to drop duplicates I used the solution here:
    https://stackoverflow.com/questions/43855462/pandas-drop-duplicates-method-not-working-on-dataframe-containing-lists

    Parameters
    ----------
    mth5_paths: list
        paths or strings that point to mth5s
    summary_type: string
        one of ["channel", "run"]
        "channel" returns concatenated channel summary,
        "run" returns concatenated run summary,
    deduplicate: bool
        Default is True, deduplicates the summary_df

    Returns
    -------
    super_summary: pd.DataFrame

    """
    dfs = len(mth5_list) * [None]

    for i, mth5_elt in enumerate(mth5_list):
        if isinstance(mth5_elt, mth5.mth5.MTH5):
            mth5_obj = mth5_elt
        else:  # mth5_elt is a path or a string
            mth5_obj = initialize_mth5(mth5_elt, mode="a")

        df = extract_run_summary_from_mth5(mth5_obj, summary_type=summary_type)

        # close it back up if you opened it
        if not isinstance(mth5_elt, mth5.mth5.MTH5):
            mth5_obj.close_mth5()
        dfs[i] = df

    # merge all summaries into a super_summary
    super_summary = pd.concat(dfs)
    super_summary.reset_index(drop=True, inplace=True)

    # drop rows that correspond to TFs:
    run_rows = super_summary.sample_rate != 0
    super_summary = super_summary[run_rows]
    super_summary.reset_index(drop=True, inplace=True)

    if deduplicate:
        keep_indices = super_summary.astype(str).drop_duplicates().index
        super_summary = super_summary.loc[keep_indices]
        super_summary.reset_index(drop=True, inplace=True)
    return super_summary



import logging
import pathlib
import unittest

from aurora.test_utils.synthetic.make_mth5_from_asc import create_test12rr_h5
from aurora.transfer_function.kernel_dataset import KernelDataset


class TestKernelDataset(unittest.TestCase):
    """ """
    @classmethod
    def setUpClass(self):
        self._mth5_path = create_test12rr_h5()
        self._kd = KernelDataset()
        self._kd.from_run_summary([self._mth5_path])

    def setUp(self):
        self._kd = KernelDataset()
        print("OK")

    def test_dot_dict(self):
        assert True

def main():
    # tmp = TestMetadataValuesSetCorrect()
    # tmp.setUp()
    # tmp.test_start_times_correct()
    unittest.main()


if __name__ == "__main__":
    main()




from aurora.pipelines.process_mth5 import process_mth5
from aurora.pipelines.run_summary import RunSummary
from aurora.sandbox.io_helpers.zfile_murphy import read_z_file
from aurora.test_utils.synthetic.make_mth5_from_asc import create_test1_h5
from aurora.test_utils.synthetic.make_mth5_from_asc import create_test2_h5
from aurora.test_utils.synthetic.make_mth5_from_asc import create_test12rr_h5
from aurora.test_utils.synthetic.make_processing_configs import (
    create_test_run_config,
)
from aurora.test_utils.synthetic.paths import AURORA_RESULTS_PATH
from aurora.test_utils.synthetic.paths import EMTF_OUTPUT_PATH
from aurora.test_utils.synthetic.rms_helpers import assert_rms_misfit_ok
from aurora.test_utils.synthetic.rms_helpers import compute_rms
from aurora.test_utils.synthetic.rms_helpers import get_expected_rms_misfit
from aurora.transfer_function.emtf_z_file_helpers import (
    merge_tf_collection_to_match_z_file,
)
from aurora.transfer_function.kernel_dataset import KernelDataset
from plot_helpers_synthetic import plot_rho_phi
from mth5.helpers import close_open_files


def aurora_vs_emtf(
    test_case_id,
    emtf_version,
    auxilliary_z_file,
    z_file_base,
    tfk_dataset,
    make_rho_phi_plot=True,
    show_rho_phi_plot=False,
    use_subtitle=True,
):
    """

    ToDo: Consider storing the processing config for this case as a json file,
    committed with the code.

    Just like a normal test of processing synthetic data, but this uses a
    known processing configuration and has a known result.  The results are plotted and
    stored and checked against a standard result calculated originally in August 2021.

    There are two cases of comparisons here.  In one case we compare against
    the committed .zss file in the EMTF repository, and in the other case we compare
    against a committed .mat file created by the matlab codes.

    Note that the comparison values got slightly worse since the original commit.
    It turns out that we can recover the original values by setting beta to the old
    formula, where beta is .8843, not .7769.

    Parameters
    ----------
    test_case_id: str
        one of ["test1", "test2r1"].  "test1" is associated with single station
        processing. "test2r1" is remote refernce processing
    emtf_version: str
        one of ["fortran", "matlab"]
    auxilliary_z_file: str or pathlib.Path
        points to a .zss, .zrr or .zmm that EMTF produced that will be compared
        against the python aurora output
    z_file_base: str
        This is the z_file that aurora will write its output to
    tfk_dataset: aurora.transfer_function.kernel_dataset.KernelDataset
        Info about the data to process
    make_rho_phi_plot: bool
    show_rho_phi_plot: bool
    use_subtitle: bool
    """
    processing_config = create_test_run_config(
        test_case_id, tfk_dataset, matlab_or_fortran=emtf_version
    )

    expected_rms_misfit = get_expected_rms_misfit(test_case_id, emtf_version)
    z_file_path = AURORA_RESULTS_PATH.joinpath(z_file_base)

    tf_collection = process_mth5(
        processing_config,
        tfk_dataset=tfk_dataset,
        z_file_path=z_file_path,
        return_collection=True,
    )

    aux_data = read_z_file(auxilliary_z_file)
    aurora_rho_phi = merge_tf_collection_to_match_z_file(aux_data, tf_collection)

    for xy_or_yx in ["xy", "yx"]:
        aurora_rho = aurora_rho_phi["rho"][xy_or_yx]
        aurora_phi = aurora_rho_phi["phi"][xy_or_yx]
        aux_rho = aux_data.rho(xy_or_yx)
        aux_phi = aux_data.phi(xy_or_yx)
        rho_rms_aurora, phi_rms_aurora = compute_rms(
            aurora_rho, aurora_phi, verbose=True
        )
        rho_rms_emtf, phi_rms_emtf = compute_rms(aux_rho, aux_phi)

        if expected_rms_misfit is not None:
            assert_rms_misfit_ok(
                expected_rms_misfit, xy_or_yx, rho_rms_aurora, phi_rms_aurora
            )

        if make_rho_phi_plot:
            plot_rho_phi(
                xy_or_yx,
                tf_collection,
                rho_rms_aurora,
                rho_rms_emtf,
                phi_rms_aurora,
                phi_rms_emtf,
                emtf_version,
                aux_data=aux_data,
                use_subtitle=use_subtitle,
                show_plot=show_rho_phi_plot,
                output_path=AURORA_RESULTS_PATH,
            )

    return


def run_test1(emtf_version, ds_df):
    """

    Parameters
    ----------
    emtf_version : string
        "matlab", or "fortran"
    ds_df : pandas.DataFrame
        Basically a run_summary dataframe

    Returns
    -------

    """
    print(f"Test1 vs {emtf_version}")
    test_case_id = "test1"
    auxilliary_z_file = EMTF_OUTPUT_PATH.joinpath("test1.zss")
    z_file_base = f"{test_case_id}_aurora_{emtf_version}.zss"
    aurora_vs_emtf(test_case_id, emtf_version, auxilliary_z_file, z_file_base, ds_df)
    return


def run_test2r1(tfk_dataset):
    """

    Parameters
    ----------
    ds_df : pandas.DataFrame
        Basically a run_summary dataframe
    Returns
    -------

    """
    print("Test2r1")
    test_case_id = "test2r1"
    emtf_version = "fortran"
    auxilliary_z_file = EMTF_OUTPUT_PATH.joinpath("test2r1.zrr")
    z_file_base = f"{test_case_id}_aurora_{emtf_version}.zrr"
    aurora_vs_emtf(
        test_case_id, emtf_version, auxilliary_z_file, z_file_base, tfk_dataset
    )
    return


def make_mth5s(merged=True):
    """
    Returns
    -------
    mth5_paths: list of Path objs or str(Path)
    """
    if merged:
        mth5_path = create_test12rr_h5()
        mth5_paths = [
            mth5_path,
        ]
    else:
        mth5_path_1 = create_test1_h5()
        mth5_path_2 = create_test2_h5()
        mth5_paths = [mth5_path_1, mth5_path_2]
    return mth5_paths


def test_pipeline(merged=True):
    """

    Parameters
    ----------
    merged: bool
        If true, summarise two separate mth5 files and merge their run summaries
        If False, use an already-merged mth5

    Returns
    -------

    """
    close_open_files()

    mth5_paths = make_mth5s(merged=merged)
    run_summary = RunSummary()
    run_summary.from_mth5s(mth5_paths)
    tfk_dataset = KernelDataset()
    tfk_dataset.from_run_summary(run_summary, "test1")

    run_test1("fortran", tfk_dataset)
    run_test1("matlab", tfk_dataset)

    tfk_dataset = KernelDataset()
    tfk_dataset.from_run_summary(run_summary, "test2", "test1")
    run_test2r1(tfk_dataset)


def test():
    import logging

    logging.getLogger("matplotlib.font_manager").disabled = True
    logging.getLogger("matplotlib.ticker").disabled = True

    test_pipeline(merged=False)
    test_pipeline(merged=True)


def main():
    test()


if __name__ == "__main__":
    main()



"""
Players on the stage:  One or more mth5s.

Each mth5 has a mth5_obj.channel_summary dataframe which tells what data are available.
Here we use a compressed view of this df with one line per acquisition run. I've been
calling that a "run_summary".  That object could be moved to mth5, so that each mth5
has a mth5_obj.run_summary dataframe.  As of Mar 29, 2023 a RunSummary is available at the
station level in mth5, but the aurora version is still being used.  This should be merged if possible
so that aurora uses the built-in mth5 method.

The run_summary provides options for the local and possibly remote reference stations.
 Candidates for local station are the unique value in the station column.
 *It maybe that we need to groupby survey & station, for now I am considering station
 names to be unique.

 For any given candidate station, there are some integer n runs available.
 This yields 2**n - 1 possible combinations that can be processed, neglecting any
 flagging of time intervals within any run, or any joining of runs.
 (There are actually 2**n, but we ignore the empty set, so -1)

 Intuition suggests default ought to be to process n runs in n+1 configurations:
 {all runs} + each run individually.  This will give a bulk answer, and bad runs can
 be flagged by comparing them.  After an initial processing, the tfs can be reviewed
 and the problematic runs can be addressed.

The user can interact with the run_summary_df, selecting sub dataframes via querying,
and in future maybe via some GUI (or a spreadsheet).


The process looks like this:
0. Start with a list of mth5s
1. Extract channel_summaries from each mth5 and join them vertically
2. Compress to a run_summay
3. Stare at the run_summary_df & Select a station "S" to process
4. Given "S"", select a non-empty set of runs for that station
5. Select a remote reference "RR", (this is allowed to be empty)
6. Extract the sub-dataframe corresponding to you local_station acquistion_runs,
and the remote station acquition runs
7. If the remote is non-empty,
a) Drop the runs (rows) associated with the remote that DO NOT intersect with local
b) restrict the start/end times of the remote runs that DO intersect with the
local so that overlap is complete.
c) restrict start/end times of the local runs so that they DO intersect with remote
8. This is now a TFKernel Dataset Definition (ish).  Initialize a default
processing object and pass it this df:
cc = ConfigCreator()
p = cc.create_from_kernel_dataset(kernel_dataset, emtf_band_file=emtf_band_setup_file)
9. Edit the Processing Config appropriately,

ToDo: Consider supporting a default value for 'channel_scale_factors' that is None,
"""

import copy
import pandas as pd

from aurora.pipelines.run_summary import RUN_SUMMARY_COLUMNS
from mt_metadata.utils.list_dict import ListDict

# Add these to a standard, so we track add/subtract columns
KERNEL_DATASET_COLUMNS = RUN_SUMMARY_COLUMNS + [
    "channel_scale_factors",
    "duration",
    "fc",
]


class KernelDataset:
    """
    This class is intended to work with mth5-derived channel_summary or run_summary
    dataframes, that specify time series intervals.

    This class is closely related to (may actually be an extension of) RunSummary

    The main idea is to specify one or two stations, and a list of acquisition "runs"
    that can be merged into a "processing run".
    Each acquistion run can be further divided into non-overlapping chunks by specifying
    time-intervals associated with that acquistion run.  An empty iterable of
    time-intervals associated with a run is interpretted as the interval
    corresponding to the entire run.

    The time intervals can be used for several purposes but primarily:
    To specify contiguous chunks of data for:
    1. STFT, that will be made into merged FC data structures
    2. binding together into xarray time series, for eventual gap fill (and then STFT)
    3. managing and analyse the availability of reference time series

    The basic data strucutre can be represented as a table or as a tree:
    Station <-- run <-- [Intervals],

    This is described in issue #118 https://github.com/simpeg/aurora/issues/118

    Desired Properties
    a) This should be able to take a dictionary (tree) and return the tabular (
    DataFrame) representation and vice versa.
    b) When there are two stations, can apply interval intersection rules, so that
    only time intervals when both stations are acquiring data are kept

    From (a) above we can see that a simple table per station can
    represent the available data.  That table can be generated by default from
    the mth5, and intervals to exclude some data can be added as needed.

    (b) is really just the case of considering pairs of tables like (a)


    Question: To return a copy or modify in-place when querying.  Need to decide on
    standards and syntax.  Handling this in general is messy because every function
    needs to be modified.  Maybe better to use a decorator that allows for df kwarg
    to be passed, and if it is not passed the modification is done in place.
    The user who doesn't want to modify in place can work with a clone.

    """

    def __init__(self, **kwargs):
        self.df = kwargs.get("df")
        self.local_station_id = kwargs.get("local_station_id")
        self.remote_station_id = kwargs.get("remote_station_id")
        self._mini_summary_columns = [
            "survey",
            "station_id",
            "run_id",
            "start",
            "end",
            "duration",
        ]
        self.survey_metadata = {}

    def clone(self):
        return copy.deepcopy(self)

    def clone_dataframe(self):
        return copy.deepcopy(self.df)

    def from_run_summary(self, run_summary, local_station_id, remote_station_id=None):
        """

        Parameters
        ----------
        run_summary: aurora.pipelines.run_summary.RunSummary
            Summary of available data for processing from one or more stations
        local_station_id: string
            Label of the station for which an estimate will be computed
        remote_station_id: string
            Label of the remote reference station

        Returns
        -------

        """
        self.local_station_id = local_station_id
        self.remote_station_id = remote_station_id

        station_ids = [
            local_station_id,
        ]
        if remote_station_id:
            station_ids.append(remote_station_id)
        df = restrict_to_station_list(run_summary.df, station_ids, inplace=False)
        df["remote"] = False
        if remote_station_id:
            cond = df.station_id == remote_station_id
            df.remote = cond

        self.df = df
        if remote_station_id:
            self.restrict_run_intervals_to_simultaneous()
        # ADD A CHECK HERE df is non-empty
        if len(self.df) == 0:
            print("No Overlap between local and remote station data streams")
            print("Remote reference processing not a valid option")
        else:
            self._add_duration_column()
        self.df["fc"] = False

    @property
    def mini_summary(self):
        return self.df[self._mini_summary_columns]

    @property
    def print_mini_summary(self):
        print(self.mini_summary)

    @property
    def local_survey_id(self):
        survey_id = self.df.loc[~self.df.remote].survey.unique()[0]
        if survey_id in ["none"]:
            survey_id = "0"
        return survey_id

    @property
    def local_survey_metadata(self):
        return self.survey_metadata[self.local_survey_id]

    def _add_duration_column(self):
        """ """
        timedeltas = self.df.end - self.df.start
        durations = [x.total_seconds() for x in timedeltas]
        self.df["duration"] = durations
        return

    def drop_runs_shorter_than(self, duration, units="s"):
        """
        This needs to have duration refreshed before hand
        Parameters
        ----------
        duration
        units

        Returns
        -------

        """
        if units != "s":
            raise NotImplementedError
        if "duration" not in self.df.columns:
            self._add_duration_column()
        drop_cond = self.df.duration < duration
        self.df.drop(self.df[drop_cond].index, inplace=True)
        self.df.reset_index(drop=True, inplace=True)
        return

    def select_station_runs(self, station_runs_dict, keep_or_drop):
        df = select_station_runs(self.df, station_runs_dict, keep_or_drop)
        self.df = df
        return

    @property
    def is_single_station(self):
        if self.local_station_id:
            if self.remote_station_id:
                return False
            else:
                return True
        else:
            return False

    @property
    def is_remote_reference(self):
        raise NotImplementedError

    def restrict_run_intervals_to_simultaneous(self):
        """
        For each run in local_station_id we check if it has overlap with other runs

        There is room for optimiztion here

        Note that you can wind up splitting runs here.  For example, in that case where
        local is running continuously, but remote is intermittent.  Then the local
        run may break into several chunks.

        Returns
        -------

        """
        local_df = self.df[self.df.station_id == self.local_station_id]
        remote_df = self.df[self.df.station_id == self.remote_station_id]
        output_sub_runs = []
        for i_local, local_row in local_df.iterrows():
            for i_remote, remote_row in remote_df.iterrows():
                if intervals_overlap(
                    local_row.start,
                    local_row.end,
                    remote_row.start,
                    remote_row.end,
                ):
                    # print(f"OVERLAP {i_local}, {i_remote}")
                    olap_start, olap_end = overlap(
                        local_row.start,
                        local_row.end,
                        remote_row.start,
                        remote_row.end,
                    )
                    # print(
                    #     f"{olap_start} -- {olap_end}\n "
                    #     f"{(olap_end-olap_start).seconds}s\n\n"
                    # )

                    local_sub_run = local_row.copy(deep=True)
                    remote_sub_run = remote_row.copy(deep=True)
                    local_sub_run.start = olap_start
                    local_sub_run.end = olap_end
                    remote_sub_run.start = olap_start
                    remote_sub_run.end = olap_end
                    output_sub_runs.append(local_sub_run)
                    output_sub_runs.append(remote_sub_run)
                else:
                    pass
                    # print(f"NOVERLAP {i_local}, {i_remote}")
        df = pd.DataFrame(output_sub_runs)
        df = df.reset_index(drop=True)

        self.df = df
        return

    def get_station_metadata(self, local_station_id):
        """
        Helper function for archiving the TF

        Parameters
        ----------
        local_station_id: str
            The name of the local station

        Returns
        -------

        """
        # get a list of local runs:
        cond = self.df["station_id"] == local_station_id
        sub_df = self.df[cond]
        sub_df.drop_duplicates(subset="run_id", inplace=True)

        # sanity check:
        run_ids = sub_df.run_id.unique()
        assert len(run_ids) == len(sub_df)

        # iterate over these runs, packing metadata into
        station_metadata = None
        for i, row in sub_df.iterrows():
            local_run_obj = self.get_run_object(row)
            if station_metadata is None:
                station_metadata = local_run_obj.station_metadata
                station_metadata.runs = ListDict()
            run_metadata = local_run_obj.metadata
            station_metadata.add_run(run_metadata)
        return station_metadata

    @property
    def num_sample_rates(self):
        return len(self.df.sample_rate.unique())

    @property
    def sample_rate(self):
        if self.num_sample_rates != 1:
            msg = "Aurora does not yet process data from mixed sample rates"
            print(f"{msg}")
            raise NotImplementedError(msg)
        sample_rate = self.df.sample_rate.unique()[0]
        return sample_rate

    def initialize_dataframe_for_processing(self, mth5_objs):
        """
        Adds extra columns needed for processing, populates them with mth5 objects,
        run_reference, and xr.Datasets.

        Note #1: When assigning xarrays to dataframe cells, df dislikes xr.Dataset,
        so we convert to xr.DataArray before packing df

        Note #2: [OPTIMIZATION] By accesssing the run_ts and packing the "run_dataarray" column of the df with it, we
         perform a non-lazy operation, and essentially forcing the entire decimation_level=0 dataset to be
         loaded into memory.  Seeking a lazy method to handle this maybe worthwhile.  For example, using
         a df.apply() approach to initialize only ione row at a time would allow us to gernerate the FCs one
         row at a time and never ingest more than one run of data at a time ...


        Parameters
        ----------
        mth5_objs: dict,  keyed by station_id
        """

        self.add_columns_for_processing(mth5_objs)

        for i, row in self.df.iterrows():
            run_obj = row.mth5_obj.get_run(
                row.station_id, row.run_id, survey=row.survey
            )
            self.df["run_reference"].at[i] = run_obj.hdf5_group.ref

            if row.fc:
                msg = f"row {row} already has fcs prescribed by processing confg "
                msg += "-- skipping time series initialzation"
                print(msg)
            #    continue
            # the line below is not lazy, See Note #2
            run_ts = run_obj.to_runts(start=row.start, end=row.end)
            self.df["run_dataarray"].at[i] = run_ts.dataset.to_array("channel")

            # wrangle survey_metadata into kernel_dataset
            survey_id = run_ts.survey_metadata.id
            if i == 0:
                self.survey_metadata[survey_id] = run_ts.survey_metadata
            elif i > 0:
                if row.station_id in self.survey_metadata[survey_id].stations.keys():
                    self.survey_metadata[survey_id].stations[row.station_id].add_run(
                        run_ts.run_metadata
                    )
                else:
                    self.survey_metadata[survey_id].add_station(run_ts.station_metadata)
            if len(self.survey_metadata.keys()) > 1:
                raise NotImplementedError

        print("DATASET DF POPULATED")

    def add_columns_for_processing(self, mth5_objs):
        """
        Moving this into kernel_dataset from processing_pipeline

        Q: Should mth5_objs be keyed by survey-station?
        A: Yes, and ...
        since the KernelDataset dataframe will be iterated over we should probably
        write an iterator method.  This can iterate over survey-station tuples
        for multiple station processing.

        Parameters
        ----------
        mth5_objs: dict,  keyed by station_id

        """
        columns_to_add = ["run_dataarray", "stft", "run_reference"]
        mth5_obj_column = len(self.df) * [None]
        for i, station_id in enumerate(self.df["station_id"]):
            mth5_obj_column[i] = mth5_objs[station_id]
        self.df["mth5_obj"] = mth5_obj_column
        for column_name in columns_to_add:
            self.df[column_name] = None

    def get_run_object(self, index_or_row):
        """

        Parameters
        ----------
        index_or_row: integer index of df, or pd.Series object

        Returns
        -------

        """
        if isinstance(index_or_row, int):
            row = self.df.loc[index_or_row]
        else:
            row = index_or_row
        run_obj = row.mth5_obj.from_reference(row.run_reference)
        return run_obj

    def close_mths_objs(self):
        """
        Loop over all unique mth5_objs in the df and make sure they are closed

        """
        mth5_objs = self.df["mth5_obj"].unique()
        for mth5_obj in mth5_objs:
            mth5_obj.close_mth5()
        return


def restrict_to_station_list(df, station_ids, inplace=True):
    """
    Drops all rows of run_summary dataframe where station_ids are NOT in
    the provided list of station_ids.  Operates on a deepcopy of self.df if a df
    isn't provided

    Parameters
    ----------
    station_ids: str or list of strings
        These are the station ids to keep, normally local and remote
    overwrite: bool
        If True, self.df is overwritten with the reduced dataframe

    Returns
    -------
        reduced dataframe with only stations associated with the station_ids
    """
    if isinstance(station_ids, str):
        station_ids = [
            station_ids,
        ]
    if not inplace:
        df = copy.deepcopy(df)
    cond1 = ~df["station_id"].isin(station_ids)
    df.drop(df[cond1].index, inplace=True)
    df = df.reset_index(drop=True)
    return df


def select_station_runs(
    df,
    station_runs_dict,
    keep_or_drop,
    overwrite=True,
):
    """
    Drops all rows where station_id==station_id, and run_id is NOT in the provided
     list of keep_run_ids.  Operates on a deepcopy df if inplace=False
    Uncommon use case the way this is coded, because it will restrict to a single
    station processing case.  Better to use drop runs, or a dict-style input

    Note1: Logic of keep/drop
    keep where cond1 is false
    keep where cond1 & cond2 both true
    drop where cond1 is true but cond2 is false

    Parameters
    ----------
    station_runs_dict: dict
        Keys are string ids of the stations to keep
        Values are lists of string labels for run_ids to keep
    keep_or_drop: str
        If "keep": returns df with only the station_rus specified in station_runs_dict
        If "drop": returns df with station_runs_dict excised
    overwrite: bool
        If True, self.df is overwritten with the reduced dataframe

    Returns
    -------
        reduced dataframe with only run_ids provided removed.
    """

    if not overwrite:
        df = copy.deepcopy(df)
    for station_id, run_ids in station_runs_dict.items():
        if isinstance(run_ids, str):
            run_ids = [
                run_ids,
            ]
        cond1 = df["station_id"] == station_id
        cond2 = df["run_id"].isin(run_ids)
        if keep_or_drop == "keep":
            drop_df = df[cond1 & ~cond2]
        else:
            drop_df = df[cond1 & cond2]

        df.drop(drop_df.index, inplace=True)
        df.reset_index(drop=True, inplace=True)
    return df


def intervals_overlap(start1, end1, start2, end2):
    """
    https://stackoverflow.com/questions/3721249/python-date-interval-intersection

    Parameters
    ----------
    start1: pd.Timestamp
        Start of interval 1
    end1: pd.Timestamp
        End of interval 1
    start2: pd.Timestamp
        Start of interval 2
    end2: pd.Timestamp
        End of interval 2

    Returns
    -------
    cond: bool
        True of the intervals overlap, False if they do now

    """
    cond = (start1 <= start2 <= end1) or (start2 <= start1 <= end2)
    return cond


def overlap(t1start, t1end, t2start, t2end):
    """
    https://stackoverflow.com/questions/3721249/python-date-interval-intersection

    Parameters
    ----------
    t1start
    t1end
    t2start
    t2end

    Returns
    -------

    """
    if t1start <= t2start <= t2end <= t1end:
        return t2start, t2end
    elif t1start <= t2start <= t1end:
        return t2start, t1end
    elif t1start <= t2end <= t1end:
        return t1start, t2end
    elif t2start <= t1start <= t1end <= t2end:
        return t1start, t1end
    else:
        return None, None


def main():
    return


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Extend the Processing class with some aurora-specific methods
"""
# =============================================================================
# Imports
# =============================================================================
import pandas as pd

from aurora.time_series.windowing_scheme import window_scheme_from_decimation
from mt_metadata.transfer_functions.processing.aurora.processing import Processing
from mt_metadata.utils.list_dict import ListDict


class Processing(Processing):
    def __init__(self, **kwargs):

        # super().__init__(attr_dict=attr_dict, **kwargs)
        super().__init__(**kwargs)

    def window_scheme(self, as_type="df"):
        """
        Make a dataframe of processing parameters one row per decimation level.
        Returns
        -------

        """
        window_schemes = [window_scheme_from_decimation(x) for x in self.decimations]
        data_dict = {}
        data_dict["sample_rate"] = [x.sample_rate for x in window_schemes]
        data_dict["window_duration"] = [x.window_duration for x in window_schemes]
        data_dict["num_samples_window"] = [x.num_samples_window for x in window_schemes]
        data_dict["num_samples_overlap"] = [
            x.num_samples_overlap for x in window_schemes
        ]
        data_dict["num_samples_advance"] = [
            x.num_samples_advance for x in window_schemes
        ]
        if as_type == "dict":
            return data_dict
        elif as_type == "df":
            df = pd.DataFrame(data=data_dict)
            return df
        else:
            print(f"unexpected rtype for window_scheme {as_type}")
            raise TypeError

    def decimation_info(self):
        decimation_ids = [x.decimation.level for x in self.decimations]
        decimation_factors = [x.decimation.factor for x in self.decimations]
        decimation_info = dict(zip(decimation_ids, decimation_factors))
        return decimation_info

    def save_as_json(self, filename=None, nested=True, required=False):
        if filename is None:
            filename = self.json_fn()
        json_str = self.to_json(nested=nested, required=required)
        with open(filename, "w") as f:
            f.write(json_str)

    def emtf_tf_header(self, dec_level_id):
        """

        Parameters
        ----------
        dec_level_id: int
            This may tolerate strings in the future, but keep as int for now

        Returns
        -------
        tfh: mt_metadata.transfer_functions.processing.aurora.transfer_function_header.TransferFunctionHeader
        """
        tfh = ListDict()
        tfh.processing_scheme = self.decimations[dec_level_id].estimator.engine
        tfh.local_station = self.stations.local
        tfh.remote_station = self.stations.remote
        tfh.input_channels = self.decimations[dec_level_id].input_channels
        tfh.output_channels = self.decimations[dec_level_id].output_channels
        tfh.reference_channels = self.decimations[dec_level_id].reference_channels
        tfh.decimation_level_id = dec_level_id

        return tfh

    def make_tf_level(self, dec_level_id):
        """
        Initialize container for a single decimation level -- "flat" transfer function.

        Parameters
        ----------
        dec_level_id: int
            This may tolerate strings in the future, but keep as int for now

        Returns
        -------
        tf_obj: aurora.transfer_function.TTFZ.TTFZ
        """
        # from aurora.transfer_function.base import TransferFunction
        from aurora.transfer_function.TTFZ import TTFZ

        tf_obj = TTFZ(
            dec_level_id,
            self.decimations[dec_level_id].frequency_bands_obj(),
            processing_config=self,
        )

        return tf_obj


class EMTFTFHeader(ListDict):
    """
    Convenince class for storing metadata for a TF estimate.
    Based on Gary Egbert's TFHeader.m originally in
    iris_mt_scratch/egbert_codes-20210121T193218Z-001/egbert_codes/matlabPrototype_10-13-20/TF/classes

    It completely depends on the Processing class
    """

    def __init__(self, **kwargs):
        """
        Parameters
        _local_station : mt_metadata.transfer_functions.tf.station.Station()
            Station metadata object for the station to be estimated (
            location, channel_azimuths, etc.)
        _remote_station: same object type as local station
            if no remote reference then this can be None
        output_channels: list
            Probably a list of channel keys -- usually ["ex","ey","hz"]
        input_channels : list
            Probably a list of channel keys -- usually ["hx","hy"]
            These are the channels being provided as input to the regression.
        reference_channels : list
            These are the channels being used from the RR station. This is a
            channel list -- usually ["hx", "hy"]
        processing_scheme: str
            Denotes the regression engine used to estimate the transfer
            function.  One of "OLS" or "RME", "RME_RR.  Future
            versions could include , "multivariate array", "multiple remote",
            etc.

        """
        super().__init__()
        self.processing_scheme = kwargs.get("processing_scheme", None)
        self.local_station = kwargs.get("local_station", None)
        self.remote_station = kwargs.get("remote_station", None)
        self.input_channels = kwargs.get("input_channels", ["hx", "hy"])
        self.output_channels = kwargs.get("output_channels", ["ex", "ey"])
        self.reference_channels = kwargs.get("reference_channels", [])
        self.decimation_level_id = kwargs.get("decimation_level_id", None)
        self.user_meta_data = None  # placeholder for anything



import numpy as np

from aurora.time_series.frequency_band_helpers import extract_band
from aurora.time_series.xarray_helpers import handle_nan
from aurora.transfer_function.regression.iter_control import IterControl
from aurora.transfer_function.regression.TRME import TRME
from aurora.transfer_function.regression.TRME_RR import TRME_RR

from aurora.transfer_function.regression.base import RegressionEstimator
from aurora.transfer_function.weights.edf_weights import (
    effective_degrees_of_freedom_weights,
)

ESTIMATOR_LIBRARY = {"OLS": RegressionEstimator, "RME": TRME, "RME_RR": TRME_RR}


def get_estimator_class(estimation_engine):
    """

    Parameters
    ----------
    estimation_engine: str
        One of the keys in the ESTIMATOR_LIBRARY, designates the method that will be
        used to estimate the transfer function

    Returns
    -------
    estimator_class: aurora.transfer_function.regression.base.RegressionEstimator
        The class that will do the TF estimation
    """
    try:
        estimator_class = ESTIMATOR_LIBRARY[estimation_engine]
    except KeyError:
        print(f"processing_scheme {estimation_engine} not supported")
        print(f"processing_scheme must be one of {list(ESTIMATOR_LIBRARY.keys())}")
        raise Exception
    return estimator_class


def set_up_iter_control(config):
    """
    TODO: Review: maybe better to just make this the __init__ method of the
    IterControl object, iter_control = IterControl(config)


    Parameters
    ----------
    config: mt_metadata.transfer_functions.processing.aurora.decimation_level.DecimationLevel

    Returns
    -------
    iter_control: aurora.transfer_function.regression.iter_control.IterControl
        Object with parameters about iteration control in regression
    """
    if config.estimator.engine in ["RME", "RME_RR"]:
        iter_control = IterControl(
            max_number_of_iterations=config.regression.max_iterations,
            max_number_of_redescending_iterations=config.regression.max_redescending_iterations,
        )
    elif config.estimator.engine in [
        "OLS",
    ]:
        iter_control = None
    return iter_control


def check_time_axes_synched(X, Y):
    """
    Utility function for checking that time axes agree

    Parameters
    ----------
    X : xarray
    Y : xarray

    Returns
    -------

    """
    """
    It is critical that X, Y, RR have the same time axes here

    Returns
    -------

    """
    if (X.time == Y.time).all():
        pass
    else:
        print("WARNING - NAN Handling could fail if X,Y dont share time axes")
        raise Exception
    return


def get_band_for_tf_estimate(
    band, config, i_dec_level, local_stft_obj, remote_stft_obj
):
    """
    Get data for TF estimation for a particular band.

    Parameters
    ----------
    band : mt_metadata.transfer_functions.processing.aurora.FrequencyBands
        object with lower_bound and upper_bound to tell stft object which
        subarray to return
    config : mt_metadata.transfer_functions.processing.aurora.decimation_level.DecimationLevel
        information about the input and output channels needed for TF
        estimation problem setup
    local_stft_obj : xarray.core.dataset.Dataset or None
        Time series of Fourier coefficients for the station whose TF is to be
        estimated
    remote_stft_obj : xarray.core.dataset.Dataset or None
        Time series of Fourier coefficients for the remote reference station

    Returns
    -------
    X, Y, RR : xarray.core.dataset.Dataset or None
        data structures as local_stft_object and remote_stft_object, but
        restricted only to input_channels, output_channels,
        reference_channels and also the frequency axes are restricted to
        being within the frequency band given as an input argument.
    """
    dec_level_config = config.decimations[0]
    print(f"Processing band {band.center_period:.6f}s")
    band_dataset = extract_band(band, local_stft_obj)
    X = band_dataset[dec_level_config.input_channels]
    Y = band_dataset[dec_level_config.output_channels]
    check_time_axes_synched(X, Y)
    if config.stations.remote:
        band_dataset = extract_band(band, remote_stft_obj)
        RR = band_dataset[dec_level_config.reference_channels]
        check_time_axes_synched(Y, RR)
    else:
        RR = None

    return X, Y, RR


def select_channel(xrda, channel_label):
    """
    Extra helper function to make process_transfer_functions more readable without
    black forcing multiline
    Parameters
    ----------
    xrda
    channel_label

    Returns
    -------

    """
    ch = xrda.sel(
        channel=[
            channel_label,
        ]
    )
    return ch


def process_transfer_functions(
    config,
    i_dec_level,
    local_stft_obj,
    remote_stft_obj,
    transfer_function_obj,
    segment_weights=None,
    channel_weights=None,
):
    """
    This method based on TTFestBand.m

    Parameters
    ----------
    config
    local_stft_obj
    remote_stft_obj
    transfer_function_obj: aurora.transfer_function.TTFZ.TTFZ
        The transfer function container ready to receive values in this method.
    segment_weights : numpy array or None
        1D array which should be of the same length as the time axis of the STFT objects
        If these weights are zero anywhere, we drop all that segment across all channels
    channel_weights : numpy array or None


    TODO:
    1. Review the advantages of excuting the regression all at once vs
    channel-by-channel.  If there is not disadvantage to always
    using a channel-by-channel approach we can modify this to only support that
    method.  However, we still need a way to get residual covariances (see issue #87)
    2. Consider push the nan-handling into the band extraction as a
    kwarg.
    3. The reindexing of the band may be done in the extraction as well.  This would
    result in an "edf-weighting-scheme-ready" format.

    Returns
    -------

    """
    # PUT COHERENCE SORTING HERE IF WIDE BAND?
    dec_level_config = config.decimations[i_dec_level]
    estimator_class = get_estimator_class(dec_level_config.estimator.engine)
    for band in transfer_function_obj.frequency_bands.bands():
        iter_control = set_up_iter_control(dec_level_config)
        X, Y, RR = get_band_for_tf_estimate(
            band, config, i_dec_level, local_stft_obj, remote_stft_obj
        )
        # if there are segment weights apply them here
        # if there are channel weights apply them here
        # Reshape to 2d - maybe push this into extract band method
        X = X.stack(observation=("frequency", "time"))
        Y = Y.stack(observation=("frequency", "time"))
        if RR is not None:
            RR = RR.stack(observation=("frequency", "time"))

        W = effective_degrees_of_freedom_weights(X, RR, edf_obj=None)
        W[W == 0] = np.nan  # use this to drop values in the handle_nan
        # apply weights
        X *= W
        Y *= W
        if RR is not None:
            RR *= W
        X = X.dropna(dim="observation")
        Y = Y.dropna(dim="observation")
        if RR is not None:
            RR = RR.dropna(dim="observation")

        # INSERT COHERENCE SORTING HERE>
        # coh_type = "local"
        # if i_dec_level == 0:
        #     from aurora.transfer_function.weights.coherence_weights import compute_coherence_weights
        #     X, Y, RR = compute_coherence_weights(X,Y,RR, coh_type=coh_type)

        if dec_level_config.estimator.estimate_per_channel:
            for ch in dec_level_config.output_channels:
                Y_ch = Y[ch].to_dataset()  # keep as a dataset, maybe not needed

                X_, Y_, RR_ = handle_nan(X, Y_ch, RR, drop_dim="observation")

                # W = effective_degrees_of_freedom_weights(X_, RR_, edf_obj=None)
                # X_ *= W
                # Y_ *= W
                # if RR is not None:
                #     RR_ *= W

                regression_estimator = estimator_class(
                    X=X_, Y=Y_, Z=RR_, iter_control=iter_control
                )
                regression_estimator.estimate()
                transfer_function_obj.set_tf(regression_estimator, band.center_period)
        else:
            X, Y, RR = handle_nan(X, Y, RR, drop_dim="observation")
            regression_estimator = estimator_class(
                X=X, Y=Y, Z=RR, iter_control=iter_control
            )
            regression_estimator.estimate()
            transfer_function_obj.set_tf(regression_estimator, band.center_period)

    return transfer_function_obj



"""
Here is a test aimed at building an h5 for CAS04 (see Issue#31)

The station CAS04 was processed with several different RR stations.  An archived
image of the EMTF results can be found here:
https://user-images.githubusercontent.com/8312937/142910549-52de8007-62bf-407c-816d-a81bed08f298.png

The reference stations were: {CAV07, NVR11, REV06}.

This test is also related to several other issues and testing out several
functionalities, including:
- make_mth5_from_fdsnclient() method of MakeMTH5.

First, an mth5 is created that summarizes available data.

This is done via querying the IRIS metadata, but could be done starting with a
stationxml file in theory, and an example showing that workflow should be added in
future.

ToDo: Test make_all_stations() with mth5_version 0.1.0 and 0.2.0
ToDo: July 2023: Identify if CAV07 is the reason for the problem ... add an option to make individual stations
CAV07 is the reason the build failed, CAV07 fails metadata pull with IndexError: index 0 is out of bounds for axis 0 with size 0


"""

import pandas as pd

from aurora.general_helper_functions import TEST_PATH
from aurora.general_helper_functions import execute_subprocess
from aurora.sandbox.mth5_channel_summary_helpers import channel_summary_to_make_mth5
from aurora.sandbox.mth5_helpers import build_request_df

from mt_metadata.timeseries.stationxml import XMLInventoryMTExperiment
from mth5.clients import FDSN
from mth5.clients.make_mth5 import MakeMTH5
from mth5.utils.helpers import initialize_mth5
from mth5.utils.helpers import read_back_data
from helper_functions import xml_to_mth5

# Define paths
CAS04_PATH = TEST_PATH.joinpath("cas04")
DATA_PATH = CAS04_PATH.joinpath("data")
DATA_PATH.mkdir(exist_ok=True)
XML_PATH = CAS04_PATH.joinpath("cas04_from_tim_20211203.xml")

# Define args for data getter
NETWORK_ID = "8P"
STATION_IDS = [
    "CAS04",
    "NVR11",
    "REV06",
]  # "CAV07",] # skip CAV07 until exception fixed in mth5/metadata
CHANNELS = [
    "LQE",
    "LQN",
    "LFE",
    "LFN",
    "LFZ",
]
START = "2020-06-02T19:00:00"
END = "2020-07-13T19:00:00"
# Test use very large time interval to build all (passes)
# START = "2000-06-02T19:00:00"
# END = "2023-07-13T19:00:00"


def make_merged_request_dataframe(
    station_ids=STATION_IDS,
    network_id=NETWORK_ID,
    channels=CHANNELS,
    start=START,
    end=END,
):
    """
    Consider moving this to sandbox
    Args:
        station_ids:
        network_id:
        channels:
        start:
        end:

    Returns:

    """
    df_list = []
    for station_id in station_ids:
        df = build_request_df(
            network_id, station_id, channels=channels, start=start, end=end
        )
        df_list.append(df)
    output_df = pd.concat(df_list)
    output_df.reset_index(inplace=True, drop=True)
    return output_df


def make_all_stations_individually(
    mth5_version="0.1.0",
):
    """
    Makes 1 h5 for each station in STATION_IDS
    Args:
        mth5_version:
    """
    for station_id in STATION_IDS:
        # request_df = build_request_df(NETWORK_ID, station_id, channels=["*F*", "*Q*", ], start=None, end=None)
        request_df = build_request_df(
            NETWORK_ID, station_id, channels=CHANNELS, start=None, end=None
        )
        fdsn_object = FDSN(mth5_version=mth5_version)
        fdsn_object.client = "IRIS"
        fdsn_object.make_mth5_from_fdsn_client(
            request_df, interact=False, path=DATA_PATH
        )
    return


def make_all_stations_together(
    mth5_version="0.1.0", return_obj=False, force_download=False
):
    """

    Parameters
    ----------
    mth5_version
    return_obj

    Returns
    -------

    """
    request_df = make_merged_request_dataframe()
    fdsn_object = FDSN(mth5_version=mth5_version)
    fdsn_object.client = "IRIS"

    expected_file_name = DATA_PATH.joinpath(fdsn_object.make_filename(request_df))
    download = force_download
    if expected_file_name.exists():
        print(f"Already have data for {expected_file_name.name}")
        download = False
        mth5_filename = expected_file_name
        if force_download:
            download = True
    if download:
        print("getting...", request_df)
        mth5_filename = fdsn_object.make_mth5_from_fdsn_client(
            request_df, interact=False, path=DATA_PATH
        )

    return mth5_filename


def test_make_mth5(mth5_version="0.1.0"):
    """

    Returns
    -------
    mth5_path: string
        Where the built mth5 lives
    """
    # make_all_stations_individually()
    mth5_path = make_all_stations_together(
        mth5_version=mth5_version, force_download=True
    )
    if mth5_version == "0.1.0":
        new_filepath = str(mth5_path).replace(".h5", "_v1.h5")
    elif mth5_version == "0.2.0":
        new_filepath = str(mth5_path).replace(".h5", "_v2.h5")
    cmd = f"mv {mth5_path} {new_filepath}"
    execute_subprocess(cmd)

    if mth5_version == "0.1.0":
        survey = None
    else:
        survey = "CONUS South"
    read_back_data(new_filepath, "CAS04", "a", survey=survey)
    read_back_data(new_filepath, "CAS04", "b", survey=survey)
    read_back_data(new_filepath, "CAS04", "c", survey=survey)
    read_back_data(new_filepath, "CAS04", "d", survey=survey)

    return new_filepath


def main():
    mth5_path = test_make_mth5(mth5_version="0.1.0")  # passes
    mth5_path = test_make_mth5(mth5_version="0.2.0")  # passes 29 Jul 2022
    return mth5_path


if __name__ == "__main__":
    main()
    print("OK")



"""
helper functions being factored out of plotter to support mulitple plots on a single
axis.
"""


def plot_rho(
    ax, periods, rho, marker="o", color="k", linestyle="None", label="", markersize=10
):
    ax.loglog(
        periods,
        rho,
        marker=marker,
        color=color,
        linestyle=linestyle,
        label=label,
        markersize=markersize,
    )
    ax.tick_params(axis="both", which="major", labelsize=16)
    ax.tick_params(axis="x", which="minor", bottom=True)
    return


def plot_phi(
    ax, periods, phi, marker="o", color="k", linestyle="None", label="", markersize=10
):
    ax.semilogx(
        periods,
        phi,
        marker=marker,
        color=color,
        linestyle=linestyle,
        label=label,
        markersize=markersize,
    )
    ax.tick_params(axis="both", which="major", labelsize=16)
    ax.minorticks_on()  # (axis="x", which="minor", bottom=True)
    return



import numpy as np


def err_log(x, y, yerr, ll, lims):
    """
    err_log : used for plotting error bars with a y-axis log scale
    takes VECTORS x and y and outputs matrices (one row per data point) for
    plotting error bars ll = 'XLOG' for log X axis

    Parameters
    ----------
    x : numpy vector
    y : numpy vector
    yerr
    ll : str
        'XLOG' for log X axis
    lims

    Returns
    -------
    [xb,yb]: matrices, one row per data point
    """
    num_observations = len(x)
    xb = np.zeros((6, num_observations))
    yb = np.zeros((6, num_observations))
    barsize = 0.0075
    if ll.lower() == "xlog":
        dx = np.log(lims[1] / lims[0]) * barsize  # natural log in matlab & python
        xb[2, :] = np.log(x)
    else:
        dx = (lims[1] - lims[0]) * barsize
        xb[2, :] = x
    xb[3, :] = xb[2, :]
    xb[0, :] = xb[2, :] - dx
    xb[1, :] = xb[2, :] + dx
    xb[4, :] = xb[2, :] - dx
    xb[5, :] = xb[2, :] + dx
    if ll.lower() == "xlog":
        xb = np.exp(xb)

    yb[0, :] = (y - yerr).T
    yb[1, :] = (y - yerr).T
    yb[2, :] = (y - yerr).T
    yb[3, :] = (y + yerr).T
    yb[4, :] = (y + yerr).T
    yb[5, :] = (y + yerr).T

    return xb, yb



from aurora.time_series.windowing_scheme import WindowingScheme
from mth5.mth5 import MTH5
from aurora.test_utils.parkfield.calibration_helpers import (
    parkfield_sanity_check,
)
from aurora.test_utils.parkfield.make_parkfield_mth5 import ensure_h5_exists
from aurora.test_utils.parkfield.path_helpers import AURORA_RESULTS_PATH
from aurora.test_utils.parkfield.path_helpers import DATA_PATH


def validate_bulk_spectra_have_correct_units(run_obj, run_ts_obj, show_spectra=False):
    """

    Parameters
    ----------
    run_obj: mth5.groups.master_station_run_channel.RunGroup
        /Survey/Stations/PKD/001:
        ====================
            --> Dataset: ex
            .................
            --> Dataset: ey
            .................
            --> Dataset: hx
            .................
            --> Dataset: hy
            .................
            --> Dataset: hz
            .................
    run_ts_obj: mth5.timeseries.run_ts.RunTS
        RunTS Summary:
            Station:     PKD
            Run:         001
            Start:       2004-09-28T00:00:00+00:00
            End:         2004-09-28T01:59:59.950000+00:00
            Sample Rate: 40.0
            Components:  ['ex', 'ey', 'hx', 'hy', 'hz']
    show_spectra: bool
        controls whether plots flash to screen in parkfield_sanity_check

    Returns
    -------

    """
    windowing_scheme = WindowingScheme(
        taper_family="hamming",
        num_samples_window=run_ts_obj.dataset.time.shape[0],  # 288000
        num_samples_overlap=0,
        sample_rate=run_ts_obj.sample_rate,  # 40.0 sps
    )
    windowed_obj = windowing_scheme.apply_sliding_window(run_ts_obj.dataset)
    tapered_obj = windowing_scheme.apply_taper(windowed_obj)

    fft_obj = windowing_scheme.apply_fft(tapered_obj)
    show_response_curves = False

    parkfield_sanity_check(
        fft_obj,
        run_obj,
        figures_path=AURORA_RESULTS_PATH,
        show_response_curves=show_response_curves,
        show_spectra=show_spectra,
        include_decimation=True,
    )
    return


def test():
    import logging

    logging.getLogger("matplotlib.font_manager").disabled = True
    logging.getLogger("matplotlib.ticker").disabled = True

    run_id = "001"
    station_id = "PKD"
    h5_path = ensure_h5_exists()
    m = MTH5(file_version="0.1.0")
    m.open_mth5(h5_path, mode="r")
    run_obj = m.get_run(station_id, run_id)
    run_ts_obj = run_obj.to_runts()
    validate_bulk_spectra_have_correct_units(run_obj, run_ts_obj, show_spectra=True)
    m.close_mth5()


def main():
    test()


if __name__ == "__main__":
    main()



import numpy as np
import pandas as pd

from mt_metadata.transfer_functions.processing.aurora.band import (
    FrequencyBands,
)


def extract_band(frequency_band, fft_obj, epsilon=1e-7):
    """
    This may become a method of fft_obj, or frequency band.
    For now leave as stand alone.

    Note #1: 20230902
    drop=True does not play nice with h5py and Dataset, results in a type error.
    File "stringsource", line 2, in h5py.h5r.Reference.__reduce_cython__
    TypeError: no default __reduce__ due to non-trivial __cinit__
    However, it works OK with DataArray, so maybe use data array in general

    Parameters
    ----------
    frequency_band: aurora.interval.Interval()
        interval corresponding to a frequency band
    fft_obj: xr.DataArray
        To be replaced with an fft_obj() class in future
    epsilon: float
        Use this when you are worried about missing a frequency due to
        round off error.  This is in general not needed if we use a df/2 pad
        around true harmonics.

    Returns
    -------
    band: xr.DataArray
        The frequencies within the band passed into this function
    """
    cond1 = fft_obj.frequency >= frequency_band.lower_bound - epsilon
    cond2 = fft_obj.frequency <= frequency_band.upper_bound + epsilon
    try:
        band = fft_obj.where(cond1 & cond2, drop=True)
    except TypeError:  # see Note #1
        tmp = fft_obj.to_array()
        band = tmp.where(cond1 & cond2, drop=True)
        band = band.to_dataset("variable")
    return band


def frequency_band_edges(
    f_lower_bound, f_upper_bound, num_bands_per_decade=None, num_bands=None
):
    """
    Provides logarithmically spaced fenceposts across lowest and highest
    frequencies. This is a lot like calling logspace.  The resultant gates
    have constant  Q, i.e. deltaF/f_center=Q=constant.
    where f_center is defined geometircally, i.e. sqrt(f2*f1) is the center freq
    between f1 and f2.

    TODO: Add a linear spacing option?

    Parameters
    ----------
    f_lower_bound : float
        lowest frequency under consideration
    f_upper_bound : float
        highest frequency under consideration
    num_bands_per_decade : int (TODO test, float maybe ok also.. need to test)
        number of bands per decade
    num_bands : int
        total number of bands.  This supercedes num_bands_per_decade if supplied

    Returns
    -------
    fence_posts : array
        logarithmically spaced fenceposts acoss lowest and highest
        frequencies.  These partition the frequency domain between
        f_lower_bound and f_upper_bound
    """
    if (num_bands is None) & (num_bands_per_decade is None):
        print("Specify either number_of_bands or numnerbands_per_decade")
        raise Exception

    if num_bands is None:
        number_of_decades = np.log10(f_upper_bound / f_lower_bound)
        # The number of decades spanned (use log8 for octaves)
        num_bands = round(
            number_of_decades * num_bands_per_decade
        )  # floor or ceiling here?

    base = np.exp((1.0 / num_bands) * np.log(f_upper_bound / f_lower_bound))
    # log - NOT log10!

    print(f"base = {base}")
    bases = base * np.ones(num_bands + 1)
    print(f"bases = {bases}")
    exponents = np.linspace(0, num_bands, num_bands + 1)
    print(f"exponents = {exponents}")
    fence_posts = f_lower_bound * (bases**exponents)
    print(f"fence posts = {fence_posts}")
    return fence_posts


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
The windowing scheme defines the chunking and chopping of the time series for
the Short Time Fourier Transform.  Often referred to as a "sliding window" or
a "striding window".  Iin its most basic form it is a taper with a rule to
say how far to advance at each stride (or step).

To generate an array of data-windows from a data series we only need the
two parameters window_length (L) and window_overlap (V).  The parameter
"window_advance" (L-V) can be used in lieu of overlap.  Sliding windows are
normally described terms of overlap but it is cleaner to code in terms of
advance.

Choices L and V are usually made with some knowledge of time series sample
rate, duration, and the frequency band of interest.  In aurora because this is used
to prep for STFT, L is typically a power of 2.

In general we will need one instance of this class per decimation level,
but in practice often leave the windowing scheme the same for each decimation level.

This class is a key part of the "gateway" to frequency domain, so it has been given
a sampling_rate attribute.  While sampling rate is a property of the data, and not
the windowing scheme per se, it is good for this class to be aware of the sampling
rate.

Future modifications could involve:
- binding this class with a time series.
- Making a subclass with only L, V, and then having an extension with sample_rate


When 2D arrays are generated how should we index them?
[[ 0  1  2]
 [ 2  3  4]
 [ 4  5  6]
 [ 6  7  8]
 [ 8  9 10]
 [10 11 12]
 [12 13 14]]
In this example the rows are indexing the individual windows ... and so they
should be associated with the time of each window.  We will need to set a
standard for this.  Obvious options are center_time of window and time_of_first
sample. I prefer time_of_first sample.  This can always be transformed to
center time or another standard later.  We can call this the "window time
axis".  The columns are indexing "steps of delta-t".  The actual times are
different for every row, so it would be best to use something like
[0, dt, 2*dt] for that axis to keep it general.  We can call this the
"within-window sample time axis"


TODO: Regarding the optional time_vector input to self.apply_sliding_window()
... this current implementation takes as input numpy array data.  We need to
also allow for an xarray to be implemented. In the simplest case we would
take an xarray in and extract its "time" axis as time vector

20210529
This class is going to be modified to only accept xarray as input data.
We can force any incoming numpy arrays to be either xr.DataArray or xr.Dataset.
Similarly, output will be only xr.DataArray or xr.Dataset
"""

import copy
import numpy as np
import xarray as xr

from aurora.time_series.apodization_window import ApodizationWindow
from aurora.time_series.windowed_time_series import WindowedTimeSeries
from aurora.time_series.window_helpers import available_number_of_windows_in_array
from aurora.time_series.window_helpers import SLIDING_WINDOW_FUNCTIONS

from mt_metadata.transfer_functions.processing.aurora.decimation_level import (
    get_fft_harmonics,
)


class WindowingScheme(ApodizationWindow):
    """
    20210415: Casting window length, overlap, advance, etc. in terms of number
    of samples or "points" here as this is common signal processing the
    nomenclature.  We may provide an interface to define these things in terms
    of percent, duration in seconds etc. in a supporting module.

    Note that sample_rate is actually a property of the data and not of the
    window ... still not sure if we want to make sample_rate an attr here
    or if its better to put properties like window_duration() as a method of
    some composition of time series and windowing scheme.
    """

    def __init__(self, **kwargs):
        super(WindowingScheme, self).__init__(**kwargs)
        self.num_samples_overlap = kwargs.get(
            "num_samples_overlap", None
        )  # make this 75% of num_samples_window by default
        self.striding_function_label = kwargs.get("striding_function_label", "crude")
        self._left_hand_window_edge_indices = None
        self.sample_rate = kwargs.get("sample_rate", None)

    def clone(cls):
        return copy.deepcopy(cls)

    def __str__(self):
        info_string = (
            f"Window of {self.num_samples_window} samples with "
            f"overlap {self.num_samples_overlap}"
        )
        # add taper summary here?
        return info_string

    @property
    def num_samples_advance(self):
        """
        A derived property.  If we made this a fundamental defined property
        then overlap would become a derived property.  Overlap is more
        conventional than advance in the literature however so we choose it as
        our property label.
        """
        return self.num_samples_window - self.num_samples_overlap

    def available_number_of_windows(self, num_samples_data):
        """

        Parameters
        ----------
        num_samples_data : int
            The number of samples in the time series to be windowed by self.

        Returns
        -------
        number_of_windows : int
           Count of the number of windows returned from time series of
           num_samples_data.  Only take as many windows as available without
           wrapping.  Start with one window for free, move forward by
           num_samples_advance and don't walk over the cliff.
        """
        return available_number_of_windows_in_array(
            num_samples_data, self.num_samples_window, self.num_samples_advance
        )

    def apply_sliding_window(
        self, data, time_vector=None, dt=None, return_xarray=False
    ):
        """
        I would like this method to support numpy arrays as well as xarrays.

        Parameters
        ----------
        data: 1D numpy array, xr.DataArray, xr.Dataset
            The data to break into ensembles.
        time_vector: 1D numpy array
            The time axis of the data.
        dt: float
            The sample interval of the data (reciprocal of sample_rate)
        return_xarray: boolean
            If True will return an xarray object, even if the input object was a
            numpy array

        Returns
        -------
        windowed_obj: arraylike
            Normally an object of type xarray.core.dataarray.DataArray
            Could be numpy array as well.
        """
        if isinstance(data, np.ndarray):
            windowed_obj = self._apply_sliding_window_numpy(
                data, time_vector=time_vector, dt=dt, return_xarray=return_xarray
            )

        elif isinstance(data, xr.DataArray):
            # Cast DataArray to DataSet, iterate and then Dataset back to DataArray
            xrds = data.to_dataset("channel")
            windowed_obj = self.apply_sliding_window(
                xrds, time_vector=time_vector, dt=dt
            )
            windowed_obj = windowed_obj.to_array("channel")

        elif isinstance(data, xr.Dataset):
            ds = xr.Dataset()
            for key in data.keys():
                windowed_obj = self._apply_sliding_window_numpy(
                    data[key].data,
                    time_vector=data.time.data,
                    dt=dt,
                    return_xarray=True,
                )
                ds.update({key: windowed_obj})
            windowed_obj = ds

        else:
            print(f"Unexpected Data type {type(data)}")
            raise Exception
        return windowed_obj

    def _apply_sliding_window_numpy(
        self, data, time_vector=None, dt=None, return_xarray=False
    ):
        """

        Parameters
        ----------
        data: numpy.ndarray
            A channel of time series data
        time_vector: numpy.ndarray or None
            Time coordinate of xarray.  If None is passed we just assign integer counts
        dt: float or None
            Sampling interval
        return_xarray: bool
            If True an xarray is returned,
            If False we just return a numpy array of the windowed data


        Returns
        -------
        output: xr.DataArray or np.ndarray
            The windowed time series, bound to time axis or just as numpy array,
            depending on the value of return_xarray
        """
        sliding_window_function = SLIDING_WINDOW_FUNCTIONS[self.striding_function_label]
        windowed_array = sliding_window_function(
            data, self.num_samples_window, self.num_samples_advance
        )

        if return_xarray:
            # Get window_time_axis coordinate
            if time_vector is None:
                time_vector = np.arange(len(data))
            window_time_axis = self.downsample_time_axis(time_vector)

            output = self.cast_windowed_data_to_xarray(
                windowed_array, window_time_axis, dt=dt
            )
        else:
            output = windowed_array

        return output

    def cast_windowed_data_to_xarray(self, windowed_array, time_vector, dt=None):
        """
        TODO?: Factor this method to a standalone function in window_helpers?

        Parameters
        ----------
        windowed_array
        time_vector
        dt

        Returns
        -------

        """
        # Get within-window_time_axis coordinate
        if dt is None:
            print("Warning dt not defined, using dt=1")
            dt = 1.0
        within_window_time_axis = dt * np.arange(self.num_samples_window)

        # cast to xr.DataArray
        xrd = xr.DataArray(
            windowed_array,
            dims=["time", "within-window time"],
            coords={"within-window time": within_window_time_axis, "time": time_vector},
        )
        return xrd

    def compute_window_edge_indices(self, num_samples_data):
        """This has been useful in the past but maybe not needed here"""
        number_of_windows = self.available_number_of_windows(num_samples_data)
        self._left_hand_window_edge_indices = (
            np.arange(number_of_windows) * self.num_samples_advance
        )
        return

    def left_hand_window_edge_indices(self, num_samples_data):
        if self._left_hand_window_edge_indices is None:
            self.compute_window_edge_indices(num_samples_data)
        return self._left_hand_window_edge_indices

    def downsample_time_axis(self, time_axis):
        """
        Parameters
        ----------
        time_axis : arraylike
            This is the time axis associated with the time-series prior to
            the windowing operation.

        Returns
        -------
        window_time_axis : array-like
            This is a time axis for the windowed data.  Say that we had 1Hz
            data starting at t=0 and 100 samples.  Then we window,
            with window length 10, and advance 10, the window time axis is
             [0, 10, 20 , ... 90].  Say the same window length, but now
             advance is 5.  Then [0, 5, 10, 15, ... 90] is the result.


        """
        lhwe = self.left_hand_window_edge_indices(len(time_axis))
        window_time_axis = time_axis[lhwe]
        return window_time_axis

    def apply_taper(self, data):
        """
        modifies the data in place by applying a taper to each window
        TODO: consider adding an option to return a copy of the data without
        the taper applied
        """
        data = WindowedTimeSeries.apply_taper(data=data, taper=self.taper)
        return data

    def frequency_axis(self, dt):
        fft_harmonics = get_fft_harmonics(self.num_samples_window, 1.0 / dt)
        return fft_harmonics

    def apply_fft(self, data, spectral_density_correction=True, detrend_type="linear"):
        """

        Parameters
        ----------
        data: xarray.core.dataset.Dataset
        spectral_density_correction: boolean
        detrend_type: string

        Returns
        -------
        spectral_ds:


        Assume we have already applied sliding window and taper.
        Things to think about:
        We want to assign the frequency axis during this method

        """
        # ONLY SUPPORTS DATASET AT THIS POINT
        if isinstance(data, xr.Dataset):
            spectral_ds = fft_xr_ds(data, self.sample_rate, detrend_type=detrend_type)
            if spectral_density_correction:
                spectral_ds = self.apply_spectral_density_calibration(spectral_ds)
        elif isinstance(data, xr.DataArray):
            xrds = data.to_dataset("channel")
            spectral_ds = fft_xr_ds(xrds, self.sample_rate, detrend_type=detrend_type)
            spectral_ds = spectral_ds.to_array("channel")
            return spectral_ds

        else:
            print(f"fft of {type(data)} not yet supported")
            raise Exception

        return spectral_ds

    def apply_spectral_density_calibration(self, dataset):
        """
        Parameters
        ----------
        dataset

        Returns
        -------


        """
        scale_factor = self.linear_spectral_density_calibration_factor
        dataset *= scale_factor
        return dataset

    # PROPERTIES THAT NEED SAMPLING RATE
    # these may be moved elsewhere later
    @property
    def dt(self):
        """
        comes from data
        """
        return 1.0 / self.sample_rate

    @property
    def window_duration(self):
        """
        units are SI seconds assuming dt is SI seconds
        """
        return self.num_samples_window * self.dt

    @property
    def duration_advance(self):
        """ """
        return self.num_samples_advance * self.dt

    @property
    def linear_spectral_density_calibration_factor(self):
        """
        Returns
        -------
        calibration_factor : float
            Following Hienzel et al 2002, Equations 24 and 25 for Linear
            Spectral Density correction for a single sided spectrum.
        """
        return np.sqrt(2 / (self.sample_rate * self.S2))


def fft_xr_ds(dataset, sample_rate, detrend_type=None, prewhitening=None):
    """

    This should call window_helpers.apply_fft_to_windowed_array
    or get moved to window_helpers.py

    The returned harmonics do not include the Nyquist frequency. To modify this
    add +1 to n_fft_harmonics.  Also, only 1-sided ffts are returned.

    For each channel within the Dataset, fft is applied along the
    within-window-time axis of the associated numpy array

    Parameters
    ----------
    dataset : xr.Dataset
        Data are 2D (windowed univariate time series).
    sample_rate: float

    detrend_type
    prewhitening

    Returns
    -------

    """
    # TODO: Modify this so that demeaning and detrending is happening before
    # application of the tapering window.  Add a second demean right before the FFT

    samples_per_window = len(dataset.coords["within-window time"])
    n_fft_harmonics = int(samples_per_window / 2)  # no bin at Nyquist,
    harmonic_frequencies = get_fft_harmonics(samples_per_window, sample_rate)

    # CORE METHOD
    output_ds = xr.Dataset()
    # operation_axis = 1  # make this pick the "time" axis from xarray
    time_coordinate_index = list(dataset.coords.keys()).index("time")
    if detrend_type:  # neither False nor None
        dataset = WindowedTimeSeries.detrend(
            data=dataset, detrend_axis=time_coordinate_index, detrend_type="linear"
        )
    for channel_id in dataset.keys():
        data = dataset[channel_id].data
        # Here is where you would add segment-by-segment prewhitening
        fspec_array = np.fft.fft(data, axis=time_coordinate_index)
        fspec_array = fspec_array[:, 0:n_fft_harmonics]  # 1-sided

        xrd = xr.DataArray(
            fspec_array,
            dims=["time", "frequency"],
            coords={"frequency": harmonic_frequencies, "time": dataset.time.data},
        )
        output_ds.update({channel_id: xrd})

    return output_ds


def window_scheme_from_decimation(decimation):
    """
    Helper function to workaround mt_metadata to not import form aurora

    Parameters
    ----------
    decimation: mt_metadata.transfer_function.processing.aurora.decimation_level
    .DecimationLevel

    Returns
    -------
        windowing_scheme aurora.time_series.windowing_scheme.WindowingScheme
    """
    from aurora.time_series.windowing_scheme import WindowingScheme

    windowing_scheme = WindowingScheme(
        taper_family=decimation.window.type,
        num_samples_window=decimation.window.num_samples,
        num_samples_overlap=decimation.window.overlap,
        taper_additional_args=decimation.window.additional_args,
        sample_rate=decimation.sample_rate_decimation,
    )
    return windowing_scheme



from matplotlib import pyplot as plt

from aurora.sandbox.io_helpers.zfile_murphy import read_z_file
from aurora.transfer_function.plot.rho_phi_helpers import plot_phi
from aurora.transfer_function.plot.rho_phi_helpers import plot_rho


def compare_two_z_files(
    z_path1,
    z_path2,
    angle1 = 0.0,
    angle2 = 0.0,
    label1="",
    label2="",
    scale_factor1=1.0,
    scale_factor2=1.0,
    out_file="",
    show_plot=True,
    use_ylims=True,
    use_xlims=True,
    **kwargs,
):
    """
    TODO: Put this in plot_helpers

    Parameters
    ----------
    z_path1: str or pathlib.Path
    z_path2: str or pathlib.Path
    angle1: float
    angle2: float
    label1: str
    label2: str
    scale_factor1
    scale_factor2
    out_file
    show_plot
    use_ylims
    use_xlims

    kwargs
    rho_ylims
    xlims


    Returns
    -------

    """
    zfile1 = read_z_file(z_path1, angle=angle1)
    zfile2 = read_z_file(z_path2, angle=angle2)
    print(f"scale_factor1: {scale_factor1}")
    fig, axs = plt.subplots(nrows=2, dpi=300, sharex=True)  # figsize=(8, 6.),
    markersize = kwargs.get("markersize", 3)

    plot_rho(
        axs[0],
        zfile1.periods,
        zfile1.rxy * scale_factor1,
        label=f"{label1} rxy",
        markersize=markersize,
        marker="^",
        color="red",
    )
    plot_rho(
        axs[0],
        zfile2.periods,
        zfile2.rxy * scale_factor2,
        label=f"{label2} rxy",
        markersize=markersize,
        marker="^",
        color="black",
    )
    plot_rho(
        axs[0],
        zfile1.periods,
        zfile1.ryx * scale_factor1,
        label=f"{label1} ryx",
        markersize=markersize,
        color="blue",
    )
    plot_rho(
        axs[0],
        zfile2.periods,
        zfile2.ryx,
        label=f"{label2} ryx",
        markersize=markersize,
        color="black",
    )
    axs[0].legend(prop={"size": 6})
    # axs[0].set_ylabel("$\rho_a$")
    axs[0].set_ylabel("Apparent Resistivity $\Omega$-m")
    rho_ylims = kwargs.get("rho_ylims", [1, 1e3])
    if use_ylims:
        axs[0].set_ylim(rho_ylims[0], rho_ylims[1])
    xlims = kwargs.get("xlims", [1e-3, 1e3])
    if use_xlims:
        axs[0].set_xlim(xlims[0], xlims[1])

    plot_phi(
        axs[1],
        zfile1.periods,
        zfile1.pxy,
        label=f"{label1} pxy",
        markersize=markersize,
        marker="^",
        color="red",
    )
    plot_phi(
        axs[1],
        zfile2.periods,
        zfile2.pxy,
        label=f"{label2} pxy",
        markersize=markersize,
        marker="^",
        color="black",
    )
    plot_phi(
        axs[1],
        zfile1.periods,
        zfile1.pyx,
        label=f"{label1} pyx",
        markersize=markersize,
        color="blue",
    )
    plot_phi(
        axs[1],
        zfile2.periods,
        zfile2.pyx,
        label=f"{label1} pyx",
        markersize=markersize,
        color="black",
    )
    axs[1].set_xlabel("Period (s)")
    axs[1].set_ylabel("Phase (degrees)")
    phi_ylims = kwargs.get("phi_ylims", [0, 90])
    axs[1].set_ylim(phi_ylims[0], phi_ylims[1])
    if out_file:
        # if out_file[-3:] != ".png":
        #     out_file+=".png"
        plt.savefig(f"{out_file}")
    if show_plot:
        plt.show()



import pandas as pd


def channel_summary_to_make_mth5(
    df, network="", channel_nomenclature=None, verbose=False
):
    """
    Context is say you have a station_xml that has come from somewhere and you want
    to make an mth5 from it, with all the relevant data.  Then you should use
    make_mth5.  But make_mth5 wants a df with a particular schema (which should be
    written down somewhere!) This method returns a dataframe with the schema that
    MakeMTH5() expects.  Specifically, there is one row for each channel-run
    combination.

    This method could be an option for output format of mth5.channel_summary()

    TODO: Add support for channel_nomenclature

    This method could be an option for output format of mth5.channel_summary()

    Parameters
    ----------
    df: pd.DataFrame
        Output from mth5_obj.channel_summary
    network: str
        Usually two characters, the network code specifies the network on which
        the data were acquired.
    verbose: bool
        Set to true to see some strings describing what is happening


    Returns
    -------
    out_df: pd.DataFrame
        The make_mth5 formatted dataframe


    """
    if not network:
        print("Network not specified")
        print("this will cause IRIS data pull to fail")
        raise ValueError

    ch_map = {"ex": "LQN", "ey": "LQE", "hx": "LFN", "hy": "LFE", "hz": "LFZ"}
    if channel_nomenclature is not None:
        print("Need to add handling for channel nomenclature")
        # either replace the keys in ch_map, or create a new ch_map

    num_rows = len(df)
    networks = num_rows * [network]
    stations = num_rows * [None]
    locations = num_rows * [""]
    channels = num_rows * [None]
    starts = num_rows * [None]
    ends = num_rows * [None]

    i = 0
    for group_id, group_df in df.groupby(["station", "run"]):
        if verbose:
            print(
                f"{group_id}, from "
                f"{group_df.start.unique()[0]}, to "
                f"{group_df.end.unique()[0]}"
            )
        for index, row in group_df.iterrows():
            stations[i] = row.station
            channels[i] = ch_map[row.component]
            starts[i] = row.start
            ends[i] = row.end
            i += 1

    out_dict = {
        "network": networks,
        "station": stations,
        "location": locations,
        "channel": channels,
        "start": starts,
        "end": ends,
    }
    out_df = pd.DataFrame(data=out_dict)
    return out_df



"""
follows Gary's TRegression.m in
iris_mt_scratch/egbert_codes-20210121T193218Z-001/egbert_codes/matlabPrototype_10-13-20/TF/classes


There are some high-level decisions to make about usage of xarray.Dataset,
xarray.DataArray and numpy arrays.  For now I am going to cast X,Y to numpy
arrays to follow Gary's codes more easily since his are matlab arrays.
"""
import numpy as np
import xarray as xr
from aurora.transfer_function.regression.iter_control import IterControl


class RegressionEstimator(object):
    """
    Abstract base class for solving Y = X*b + epsilon for b, complex-valued

    Many of the robust transfer estimation methods we will use repeat the
    model of solving Y = X*b +epsilon for b.  X is variously called the "input",
    "predictor", "explanatory", "confounding", "independent" "exogenous", variable(s)
    or the "design matrix", "model matrix" or "regressor matrix".
    Y are variously called the the "output", "predicted", "outcome",
    "response", "endogenous", "regressand", or "dependent" variable.  I will try to
    use input and output.

    When we "regress Y on X", we use the values of variable X to predict
    those of Y.

    Typically operates on single "frequency_band".
    Allows multiple columns of Y, but estimates b for each column separately.


    Estimated signal and noise covariance matrices can be used to compute error
    together to compute covariance for the matrix of regression coefficients b

    If the class has public attributes, they may be documented here
    in an ``Attributes`` section and follow the same formatting as a
    function's ``Args`` section. Alternatively, attributes may be documented
    inline with the attribute's declaration (see __init__ method below).

    Properties created with the ``@property`` decorator should be documented
    in the property's getter method.



    Attributes
    ----------
    _X : xarray.Dataset
        X.data is numpy array (normally 2-dimensional)
        These are the input variables.  Like the matlab codes each observation
        corresponds to a row and each parameter (channel) is a column.
    X :  numpy array (normally 2-dimensional)
        This is a "pure array" representation of _X used to emulate Gary
        Egbert's matlab codes. It may or may not be deprecated.
    _Y : xarray.Dataset
        These are the output variables, aranged same as X above.
    Y : numpy array (normally 2-dimensional)
        This is a "pure array" representation of _X used to emulate Gary
        Egbert's matlab codes. It may or may not be deprecated.
    b : numpy array (normally 2-dimensional)
        Matrix of regression coefficients, i.e. the solution to the regression
        problem.  In our context they are the "Transfer Function"
        The dimension of b, to match X and Y is [n_input_ch, n_output_ch]
        When we are solving "channel-by-channel", b is usually [2,1].

    inverse_signal_covariance: numpy array (n_input_ch, n_input_ch)
        This was Cov_SS in Gary's matlab codes.  It is basically inv(X.H @ X)
        Reference needed
    noise_covariance : numpy array (n_output_ch, n_output_ch)
        This was Cov_NN in Gary's matlab codes
        Reference needed
    squared_coherence: numpy array (n_output_ch)
        This was R2 in Gary's matlab codes
        Formula?  Reference?
        Squared coherence (top row is using raw data, bottom cleaned, with crude
        correction for amount of downweighted data)
    Yc : numpy array (normally 2-dimensional)
        A "cleaned" version of Y the output variables.
    iter_control : transfer_function.iter_control.IterControl()
        is a structure which controls the robust scheme
        Fields: r0, RG.nITmax, tol (rdcndwt ... not coded yet)
        On return also contains number of iterations
    """

    def __init__(self, **kwargs):
        """

        Parameters

        ----------
        kwargs
        """
        self._X = kwargs.get("X", None)
        self._Y = kwargs.get("Y", None)
        self.b = None
        self.cov_nn = None
        self.cov_ss_inv = None
        self.squared_coherence = None
        self.iter_control = kwargs.get("iter_control", IterControl())

        self.X = self._X.to_array().data.T
        self.Y = self._Y.to_array().data.T
        self.Yc = np.zeros(self.Y.shape, dtype=np.complex128)
        self.check_number_of_observations_xy_consistent()
        self.R2 = None
        self.qr_input = "X"
        self._Q = None
        self._R = None
        self._QH = None  # conjugate transpose of Q (Hermitian operator)
        self._QHY = None  #
        self._QHYc = None

    @property
    def inverse_signal_covariance(self):
        return self.cov_ss_inv

    @property
    def noise_covariance(self):
        return self.cov_nn

    def b_to_xarray(self):
        xra = xr.DataArray(
            np.transpose(self.b),
            dims=["output_channel", "input_channel"],
            coords={
                "output_channel": list(self._Y.data_vars),
                "input_channel": list(self._X.data_vars),
            },
        )
        return xra

    def solve_underdetermined(self):
        """
        20210806
        This method was originally in TRME.m, but it does not depend in
        general on using RME method so I am putting it in the base class.

        We basically never get here and when we do we dont trust the results
        https://docs.scipy.org/doc/numpy-1.9.2/reference/generated/numpy.linalg.svd.html
        https://www.mathworks.com/help/matlab/ref/double.svd.html

        <ORIGINAL MATLAB>
            <COMMENT>
        Overdetermined problem...use svd to invert, return
        NOTE: the solution IS non - unique... and by itself RME is not setup
        to do anything sensible to resolve the non - uniqueness(no prior info
        is passed!).  This is stop-gap, to prevent errors when using RME as
        part of some other estimation scheme!
            </COMMENT>
            <CODE>
        [u,s,v] = svd(obj.X,'econ');
        sInv = 1./diag(s);
        obj.b = v*diag(sInv)*u'*obj.Y;
        if ITER.returnCovariance
           obj.Cov_NN = zeros(K,K);
           obj.Cov_SS = zeros(nParam,nParam);
        end
        result = obj.b
            </CODE>
        </ORIGINAL MATLAB>


        -------

        """
        U, s, V = np.linalg.svd(self.X, full_matrices=False)
        S_inv = np.diag(1.0 / s)
        self.b = (V.T @ S_inv @ U.T) * self.Y
        if self.iter_control.return_covariance:
            print("Warning covariances are not xarray, may break things downstream")
            self.cov_nn = np.zeros((self.n_channels_out, self.n_channels_out))
            self.cov_ss_inv = np.zeros((self.n_channels_in, self.n_channels_in))

        return

    def check_number_of_observations_xy_consistent(self):
        if self.Y.shape[0] != self.X.shape[0]:
            print(
                f"Design matrix (X) has {self.X.shape[0]} rows but data (Y) "
                f"has {self.Y.shape[0]}"
            )
            raise Exception

    @property
    def n_data(self):
        """
        or return self.Y.shape[0], any reason to choose one or the other?
        See Also Issue#7 in aurora github: Masked arrays stuff will go here
        Returns
        -------

        """
        return self.X.shape[0]

    @property
    def n_channels_in(self):
        return self.X.shape[1]

    @property
    def n_channels_out(self):
        """number of output variables"""
        return self.Y.shape[1]

    @property
    def degrees_of_freedom(self):
        return self.n_data - self.n_channels_in

    @property
    def is_underdetermined(self):
        return self.degrees_of_freedom < 0

    def mask_input_channels(self):
        """
        ADD NAN MANAGEMENT HERE
        Returns
        -------

        """
        pass

    def qr_decomposition(self, X=None, sanity_check=False):
        """
        performs QR decomposition on input matrix X.  If X is not provided as a kwarg
        then check the value of self.qr_input
        Parameters
        ----------
        X: numpy array
            In TRME this is the Input channels X
            In TRME_RR this is the RR channels Z
        sanity_check: boolean
            check QR decomposition is working correctly.  Can probably be deprecated.

        Returns
        -------

        """
        if X is None:
            if self.qr_input == "X":
                X = self.X
            elif self.qr_input == "Z":
                X = self.Z
            else:
                print("Matrix to perform QR decompostion not specified")
                raise Exception

        Q, R = np.linalg.qr(X)
        self._Q = Q
        self._R = R
        if sanity_check:
            if np.isclose(np.matmul(Q, R) - X, 0).all():
                pass
            else:
                print("Failed QR decompostion sanity check")
                raise Exception
        return Q, R

    @property
    def Q(self):
        return self._Q

    @property
    def R(self):
        return self._R

    @property
    def QH(self):
        if self._QH is None:
            self._QH = self.Q.conj().T
        return self._QH

    @property
    def QHY(self):
        if self._QHY is None:
            self._QHY = self.QH @ self.Y
        return self._QHY

    # @property
    # def QHYc(self):
    #     if self._QHYc is None:
    #         self.update_QHYc()
    #     return self._QHYc
    #
    # def update_QHYc(self):
    #     self._QHYc = self.QH @ self.Yc

    def estimate_ols(self, mode="solve"):
        """

        Parameters
        ----------
        mode : str
            "qr", "brute_force", "solve"

        Returns
        -------
        b : numpy array
            Normally the impedance tensor Z

        Solve Y = Xb

        Brute Force tends to be less stable because we actually compute the
        inverse matrix.  It is not recommended, its just here for completeness.
        X'Y=X'Xb
        (X'X)^-1 X'Y = b

        -------

        """
        if mode.lower() == "qr":
            from scipy.linalg import solve_triangular

            self.qr_decomposition(self.X)
            b = solve_triangular(self.R, self.QH @ self.Y)
        else:
            X = self.X
            Y = self.Y
            XH = np.conj(X.T)
            XHX = XH @ X
            XHY = XH @ Y
            if mode.lower() == "brute_force":
                XHX_inv = np.linalg.inv(XHX)
                b = np.matmul(XHX_inv, XHY)
            elif mode.lower() == "solve":
                b = np.linalg.solve(XHX, XHY)
            else:
                print(f"mode {mode} not recognized")
                raise Exception
        self.b = b
        return b

    def estimate(self):
        Z = self.estimate_ols(mode="qr")
        return Z



"""
Supporting codes for building the FC level of the mth5


Here are the parameters that are defined via the mt_metadata fourier coefficients structures
"anti_alias_filter": "default",
"bands"
"decimation.factor": 4.0,
"decimation.level": 2,
"decimation.method": "default",
"decimation.sample_rate": 0.0625,
"extra_pre_fft_detrend_type": "linear",
"prewhitening_type": "first difference",
"window.clock_zero_type": "ignore",
"window.num_samples": 128,
"window.overlap": 32,
"window.type": "boxcar"

Key to creating the decimations config is the decision about decimation factors and the number of levels.
We have been getting this from the EMTF band setup file by default.  It is desireable to continue supporting this,
however, note that the EMTF band setup is really about processing, and not about making STFTs.

What we really want here is control of the decimation config.
This was controlled by decset.cfg which looks like this:
4     0      # of decimation level, & decimation offset
128  32.   1   0   0   7   4   32   1
1.0
128  32.   4   0   0   7   4   32   4
.2154  .1911   .1307   .0705
128  32.   4   0   0   7   4   32   4
.2154  .1911   .1307   .0705
128  32.   4   0   0   7   4   32   4
.2154  .1911   .1307   .0705

This essentially corresponds to a "Decimations Group" which is a list of decimations.
Related to the generation of FCs is the ARMA prewhitening (Issue #60) which was controlled in
EMTF with pwset.cfg
4    5             # of decimation level, # of channels
3 3 3 3 3
3 3 3 3 3
3 3 3 3 3
3 3 3 3 3

Note 1: Assumes application of cascading decimation, and that the
decimated data will be accessed from the previous decimation level.

Note 2: We can encounter cases where some runs can be decimated and others can not.
We need a way to handle this. For example, a short run may not yield any data from a
later decimation level. An attempt to handle this has been made in TF Kernel by
adding a is_valid_dataset column, associated with each run-decimation level pair.

Note 3: This point in the loop marks the interface between _generation_ of the FCs and
 their _usage_. In future the code above this comment would be pushed into
 create_fourier_coefficients() and the code below this would access those FCs and
 execute compute_transfer_function()



Questions:
1. Shouldn;t there be an experiment column in the channel_summary dataframe for a v0.2.0 file?
GROUPBY_COLUMNS = ["survey", "station", "sample_rate"]
If I use ["experiment", "survey", "station", "sample_rate"] instead (for a v0.2.0 file) encounter KeyError.

2. How to assign default values to Decimation.time_period?
Usually we will want to convert the entire run, so these should be assigned
during processing when we knwo the run extents.  Thus the
"""
# =============================================================================
# Imports
# =============================================================================


from aurora.pipelines.time_series_helpers import calibrate_stft_obj
from aurora.pipelines.time_series_helpers import prototype_decimate
from aurora.pipelines.time_series_helpers import run_ts_to_stft_scipy
from mth5.mth5 import MTH5
import mt_metadata.timeseries.time_period
from mt_metadata.transfer_functions.processing.fourier_coefficients import (
    Decimation as FCDecimation,
)


# =============================================================================
FILE_VERSION = "you need to set this, and ideally cycle over 0.1.0, 0.2.0"
DEFAULT_TIME = "1980-01-01T00:00:00+00:00"
GROUPBY_COLUMNS = ["survey", "station", "sample_rate"]  # See Question 1


def decimation_and_stft_config_creator(
    initial_sample_rate, max_levels=6, decimation_factors=None, time_period=None
):
    """
    Based on the number of samples in the run, we can compute the maximum number of valid decimation levels.
    This would re-use code in processing summary ... or we could just decimate until we cant anymore?

    You can provide soemthing like: decimation_info = {0: 1.0, 1: 4.0, 2: 4.0, 3: 4.0}

    Note 1:  This does not yet work through the assignment of which bands to keep.  Refer to
    mt_metadata.transfer_functions.processing.Processing.assign_bands() to see how this was done in the past

    Args:
        initial_sample_rate:
        max_levels:
        decimation_factors:
        time_period:

    Returns:
        decimation_and_stft_config: list
            Each element of the list is a Decimation() object.  The order of the list implies the order of the cascading
            decimation (thus no decimation levels are omitted).  This could be changed in future by using a dict
            instead of a list, e.g. decimation_factors = dict(zip(np.arange(max_levels), decimation_factors))

    """
    if not decimation_factors:
        # msg = "No decimation factors given, set default values to EMTF default values [1, 4, 4, 4, ..., 4]")
        # logger.info(msg)
        default_decimation_factor = 4
        decimation_factors = max_levels * [default_decimation_factor]
        decimation_factors[0] = 1

    # See Note 1
    decimation_and_stft_config = []
    for i_dec_level, decimation_factor in enumerate(decimation_factors):
        dd = FCDecimation()
        dd.decimation_level = i_dec_level
        dd.decimation_factor = decimation_factor
        if i_dec_level == 0:
            current_sample_rate = 1.0 * initial_sample_rate
        else:
            current_sample_rate /= decimation_factor
        dd.sample_rate_decimation = current_sample_rate

        if time_period:
            if isinstance(mt_metadata.timeseries.time_period.TimePeriod, time_period):
                dd.time_period = time_period
            else:
                print(f"Not sure how to assign time_period with {time_period}")
                raise NotImplementedError

        decimation_and_stft_config.append(dd)

    return decimation_and_stft_config


def add_fcs_to_mth5(mth5_path, decimation_and_stft_configs=None):
    """
    usssr_grouper: output of a groupby on unique {survey, station, sample_rate} tuples

    Args:
        mth5_path: str or pathlib.Path
            Where the mth5 file is locatid
        decimation_and_stft_configs:

    Returns:

    """
    m = MTH5()
    m.open_mth5(mth5_path)
    channel_summary_df = m.channel_summary.to_dataframe()

    usssr_grouper = channel_summary_df.groupby(GROUPBY_COLUMNS)
    print(f"DETECTED {len(usssr_grouper)} unique station-sample_rate instances")

    for (survey, station, sample_rate), usssr_group in usssr_grouper:
        print(f"\n\n\nsurvey: {survey}, station: {station}, sample_rate {sample_rate}")
        station_obj = m.get_station(station, survey)
        run_summary = station_obj.run_summary

        # Get the FC schemes
        if not decimation_and_stft_configs:
            msg = "FC config not supplied, using default, creating on the fly"
            print(f"{msg}")
            decimation_and_stft_configs = decimation_and_stft_config_creator(
                sample_rate, time_period=None
            )

        # Make this a function that can be done using df.apply()
        # I wonder if daskifiying that will cause issues with multiple threads trying to
        # write to the hdf5 file -- will need testing
        for i_run_row, run_row in run_summary.iterrows():
            print(
                f"survey: {survey}, station: {station}, sample_rate {sample_rate}, i_run_row {i_run_row}"
            )
            # Access Run
            run_obj = m.from_reference(run_row.hdf5_reference)

            # Set the time period:
            for decimation_and_stft_config in decimation_and_stft_configs:
                decimation_and_stft_config.time_period = run_obj.metadata.time_period

            runts = run_obj.to_runts(
                start=decimation_and_stft_config.time_period.start,
                end=decimation_and_stft_config.time_period.end,
            )
            # runts = run_obj.to_runts() # skip setting time_period explcitly

            run_xrds = runts.dataset
            # access container for FCs
            fc_group = station_obj.fourier_coefficients_group.add_fc_group(
                run_obj.metadata.id
            )

            print(" TIMING CORRECTIONS WOULD GO HERE ")

            for i_dec_level, decimation_stft_obj in enumerate(
                decimation_and_stft_configs
            ):
                if i_dec_level != 0:
                    # Apply decimation
                    run_xrds = prototype_decimate(decimation_stft_obj, run_xrds)
                print(f"type decimation_stft_obj = {type(decimation_stft_obj)}")
                if not decimation_stft_obj.is_valid_for_time_series_length(
                    run_xrds.time.shape[0]
                ):
                    print(
                        f"Decimation Level {i_dec_level} invalid, TS of {run_xrds.time.shape[0]} samples too short"
                    )
                    continue

                stft_obj = run_ts_to_stft_scipy(decimation_stft_obj, run_xrds)
                stft_obj = calibrate_stft_obj(stft_obj, run_obj)

                # print("Pack FCs into h5 and update metadata")
                decimation_level = fc_group.add_decimation_level(f"{i_dec_level}")
                decimation_level.from_xarray(stft_obj)
                decimation_level.update_metadata()
                fc_group.update_metadata()

    m.close_mth5()
    return


def read_back_fcs(mth5_path):
    """
    This is mostly a helper function for tests.  It was used as a sanity check while debugging the FC files, and
    also is a good example for how to access the data at each level for each channel.

    The Time axis of the FC array will change from level to level, but the frequency axis will stay the same shape
    (for now -- storing all fcs by default)

    Args:
        mth5_path: str or pathlib.Path
            The path to an h5 file that we will scan the fcs from

    Returns:

    """
    m = MTH5()
    m.open_mth5(mth5_path)
    channel_summary_df = m.channel_summary.to_dataframe()
    print(channel_summary_df)
    usssr_grouper = channel_summary_df.groupby(GROUPBY_COLUMNS)
    for (survey, station, sample_rate), usssr_group in usssr_grouper:
        print(f"survey: {survey}, station: {station}, sample_rate {sample_rate}")
        station_obj = m.get_station(station, survey)
        fc_groups = station_obj.fourier_coefficients_group.groups_list
        print(f"FC Groups: {fc_groups}")
        for run_id in fc_groups:
            fc_group = station_obj.fourier_coefficients_group.get_fc_group(run_id)
            dec_level_ids = fc_group.groups_list
            for dec_level_id in dec_level_ids:
                dec_level = fc_group.get_decimation_level(dec_level_id)
                print(
                    f"dec_level {dec_level_id}"
                )  # channel_summary {dec_level.channel_summary}")
                xrds = dec_level.to_xarray(["hx", "hy"])
                print(f"Time axis shape {xrds.time.data.shape}")
                print(f"Freq axis shape {xrds.frequency.data.shape}")
    return True


def main():
    pass


if __name__ == "__main__":
    main()



from obspy import UTCDateTime
from obspy.core import Stream
from obspy.core import Trace

from aurora.general_helper_functions import TEST_PATH
from aurora.test_utils.dataset_definitions import TEST_DATA_SET_CONFIGS
from aurora.test_utils.parkfield.testing_data import get_example_array_list
from aurora.sandbox.mth5_helpers import get_experiment_from_obspy_inventory
from mth5.utils.helpers import initialize_mth5
from mth5.utils.helpers import read_back_data
from mth5.timeseries import RunTS

DATA_PATH = TEST_PATH.joinpath("parkfield", "data")


def create_from_local_data(dataset_id, run_id):
    """
    TODO:
    This test could be made to depend on mth5_test_data and the imports from
    mth5_test_data can be placed within this function.
    In fact, this function could be migrated to mth5_test_data entirely,
    removing that dependency from aurora but leaving the code available for
    development and to act as an example.
    Parameters
    ----------
    dataset_id

    Returns
    -------

    """
    dataset_config = TEST_DATA_SET_CONFIGS[dataset_id]
    inventory = dataset_config.get_inventory(ensure_inventory_stages_are_named=True)
    experiment = get_experiment_from_obspy_inventory(inventory)
    run_metadata = experiment.surveys[0].stations[0].runs[0]
    station_id = "PKD"
    start_time = UTCDateTime("2004-09-28T00:00:00.000000Z")
    end_time = UTCDateTime("2004-09-28T01:59:59.975000Z")
    target_folder = DATA_PATH
    h5_path = target_folder.joinpath(f"{dataset_config.dataset_id}_from_local.h5")
    mth5_obj = initialize_mth5(h5_path)
    mth5_obj.from_experiment(experiment)
    station_group = mth5_obj.get_station(station_id)
    # </20210713>
    components_list = ["hx", "hy", "ex", "ey"]
    array_list = get_example_array_list(
        components_list=components_list, load_actual=True, station_id=station_id
    )
    header = {}
    header["sampling_rate"] = 40.0
    header["starttime"] = start_time
    header["endtime"] = end_time
    header["station"] = station_id
    header["network"] = "BK"
    # unused headers:['calib','location']

    header["channel"] = "BF1"
    hx = Trace(data=array_list[0].ts, header=header)
    header["channel"] = "BF2"
    hy = Trace(data=array_list[1].ts, header=header)
    header["channel"] = "BQ1"
    ex = Trace(data=array_list[2].ts, header=header)
    header["channel"] = "BQ2"
    ey = Trace(data=array_list[3].ts, header=header)
    run_stream = Stream(traces=[hx, hy, ex, ey])

    run_ts_obj = RunTS()
    run_ts_obj.from_obspy_stream(run_stream, run_metadata)
    run_ts_obj.run_metadata.id = run_id
    run_group = station_group.add_run(run_id)
    run_group.from_runts(run_ts_obj)

    mth5_obj.close_mth5()


def test_make_parkfield_mth5_from_local():
    dataset_id = "pkd_sao_test_00"
    run_id = "001"
    create_from_local_data(dataset_id, run_id)
    h5_path = DATA_PATH.joinpath(f"{dataset_id}_from_local.h5")
    read_back_data(h5_path, "PKD", run_id)


def main():
    test_make_parkfield_mth5_from_local()


if __name__ == "__main__":
    main()
